# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
from collections import OrderedDict

import pytest
from numpy.testing import assert_equal, assert_allclose

import MDAnalysis as mda
import os

from MDAnalysisTests.datafiles import CRD
from MDAnalysisTests import make_Universe


class TestCRDWriter(object):
    @pytest.fixture()
    def u(self):
        return mda.Universe(CRD)

    @pytest.fixture()
    def outfile(self, tmpdir):
        return os.path.join(str(tmpdir), "test.crd")

    @pytest.mark.parametrize(
        "testfile", ["test.crd", "test.crd.bz2", "test.crd.gz"]
    )
    def test_write_atoms(self, u, testfile, tmpdir):
        # Test that written file when read gives same coordinates
        with tmpdir.as_cwd():
            u.atoms.write(testfile)

            u2 = mda.Universe(testfile)

            assert_equal(u.atoms.positions, u2.atoms.positions)

    def test_roundtrip(self, u, outfile):
        # Write out a copy of the Universe, and compare this against the original
        # This is more rigorous than simply checking the coordinates as it checks
        # all formatting
        u.atoms.write(outfile)

        def CRD_iter(fn):
            with open(fn, "r") as inf:
                for line in inf:
                    if not line.startswith("*"):
                        yield line

        for ref, other in zip(CRD_iter(CRD), CRD_iter(outfile)):
            assert ref == other

    def test_write_EXT(self, u, outfile):
        # Use the `extended` keyword to force the EXT format
        u.atoms.write(outfile, extended=True)

        with open(outfile, "r") as inf:
            format_line = inf.readlines()[2]
        assert "EXT" in format_line, "EXT format expected"

    def test_write_EXT_read(self, u, outfile):
        # Read EXT format and check atom positions
        u.atoms.write(outfile, extended=True)

        u2 = mda.Universe(outfile)

        sel1 = u.select_atoms("all")
        sel2 = u2.select_atoms("all")

        cog1 = sel1.center_of_geometry()
        cog2 = sel2.center_of_geometry()

        assert_equal(
            len(u.atoms), len(u2.atoms)
        ), "Equal number of " "atoms expected in both CRD formats"
        assert_equal(
            len(u.atoms.residues), len(u2.atoms.residues)
        ), "Equal number of residues expected in" "both CRD formats"
        assert_equal(
            len(u.atoms.segments), len(u2.atoms.segments)
        ), "Equal number of segments expected in" "both CRD formats"
        assert_allclose(
            cog1, cog2, rtol=1e-6, atol=0
        ), "Same centroid expected for both CRD formats"


class TestCRDWriterMissingAttrs(object):
    # All required attributes with the default value
    req_attrs = OrderedDict(
        [
            ("resnames", "UNK"),
            ("resids", 1),
            ("names", "X"),
            ("tempfactors", 0.0),
        ]
    )

    @pytest.mark.parametrize("missing_attr", req_attrs)
    def test_warns(self, missing_attr, tmpdir):
        attrs = list(self.req_attrs.keys())
        attrs.remove(missing_attr)
        u = make_Universe(attrs, trajectory=True)

        outfile = str(tmpdir) + "/out.crd"
        with pytest.warns(UserWarning):
            u.atoms.write(outfile)

    @pytest.mark.parametrize("missing_attr", req_attrs)
    def test_write(self, missing_attr, tmpdir):
        attrs = list(self.req_attrs.keys())
        attrs.remove(missing_attr)
        u = make_Universe(attrs, trajectory=True)

        outfile = str(tmpdir) + "/out.crd"
        u.atoms.write(outfile)
        u2 = mda.Universe(outfile)

        # Check all other attrs aren't disturbed
        for attr in attrs:
            assert_equal(getattr(u.atoms, attr), getattr(u2.atoms, attr))
        # Check missing attr is as expected
        assert_equal(
            getattr(u2.atoms, missing_attr), self.req_attrs[missing_attr]
        )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""
dimensionality reduction frontend --- :mod:`MDAnalysis.analysis.encore.dimensionality_reduction.reduce_dimensionality`
======================================================================================================================

The module defines a function serving as front-end for various dimensionality
reduction algorithms, wrapping them to allow them to be used interchangably.

:Author: Matteo Tiberti, Wouter Boomsma, Tone Bengtsen

.. versionadded:: 0.16.0

.. deprecated:: 2.8.0
   This module is deprecated in favour of the 
   MDAKit `mdaencore <https://mdanalysis.org/mdaencore/>`_ and will be removed
   in MDAnalysis 3.0.0.

"""
import numpy as np
from ..confdistmatrix import get_distance_matrix
from ..utils import ParallelCalculation, merge_universes
from ..dimensionality_reduction.DimensionalityReductionMethod import (
    StochasticProximityEmbeddingNative,
)


def reduce_dimensionality(
    ensembles,
    method=None,
    select="name CA",
    distance_matrix=None,
    allow_collapsed_result=True,
    ncores=1,
    **kwargs,
):
    """
    Reduce dimensions in frames from one or more ensembles, using one or more
    dimensionality reduction methods. The function optionally takes
    pre-calculated distances matrices as an argument. Note that not all
    dimensionality reduction procedure can work directly on distance matrices,
    so the distance matrices might be ignored for particular choices of
    method.


    Parameters
    ----------

    ensembles : MDAnalysis.Universe, or list or list of list thereof
        The function takes either a single Universe object, a list of Universe
        objects or a list of lists of Universe objects. If given a single
        universe, it simply works on the conformations in the trajectory. If
        given a list of ensembles, it will merge them and analyse them together,
        keeping track of the ensemble to which each of the conformations belong.
        Finally, if passed a list of list of ensembles, the function will just
        repeat the functionality just described - merging ensembles for each
        ensemble in the outer loop.

    method : MDAnalysis.analysis.encore.dimensionality_reduction.DimensionalityReductionMethod or list
        A single or a list of instances of the DimensionalityReductionMethod
        classes from the dimensionality_reduction module. A separate analysis
        will be run for each method. Note that different parameters for the
        same method can be explored by adding different instances of
        the same dimensionality reduction class. Options are Stochastic
        Proximity Embedding or Principal Component Analysis.

    select : str, optional
        Atom selection string in the MDAnalysis format (default is "name CA")

    distance_matrix : encore.utils.TriangularMatrix, optional
        Distance matrix for stochastic proximity embedding. If this parameter
        is not supplied an RMSD distance matrix will be calculated on the fly (default).
        If several distance matrices are supplied, an analysis will be done
        for each of them. The number of provided distance matrices should
        match the number of provided ensembles.

    allow_collapsed_result: bool, optional
        Whether a return value of a list of one value should be collapsed
        into just the value (default = True).

    ncores : int, optional
        Maximum number of cores to be used (default is 1).


    Returns
    -------

    list of coordinate arrays in the reduced dimensions (or potentially a single
    coordinate array object if allow_collapsed_result is set to True)


    Example
    -------
    Two ensembles are created as Universe object using a topology file and
    two trajectories. The topology- and trajectory files used are obtained
    from the MDAnalysis test suite for two different simulations of the protein
    AdK.
    Here, we reduce two ensembles to two dimensions, and plot the result using
    matplotlib: ::

        >>> from MDAnalysis import Universe
        >>> import MDAnalysis.analysis.encore as encore
        >>> from MDAnalysis.tests.datafiles import PSF, DCD, DCD2
        >>> ens1 = Universe(PSF, DCD)
        >>> ens2 = Universe(PSF, DCD2)
        >>> coordinates, details = encore.reduce_dimensionality([ens1,ens2])
        >>> plt.scatter(coordinates[0], coordinates[1],
                        color=[["red", "blue"][m-1] for m
                        in details["ensemble_membership"]])

    Note how we extracted information about which conformation belonged to
    which ensemble from the details variable.

    You can change the parameters of the dimensionality reduction method
    by explicitly specifying the method ::

        >>> coordinates, details =
                encore.reduce_dimensionality([ens1,ens2],
                     method=encore.StochasticProximityEmbeddingNative(dimension=3))

    Here is an illustration using Principal Component Analysis, instead
    of the default dimensionality reduction method ::

        >>> coordinates, details =
                encore.reduce_dimensionality(
                     [ens1,ens2],
                     method=encore.PrincipalComponentAnalysis(dimension=2))

    You can also combine multiple methods in one call ::

        >>> coordinates, details =
                encore.reduce_dimensionality(
                     [ens1,ens2],
                     method=[encore.PrincipalComponentAnalysis(dimension=2),
                             encore.StochasticProximityEmbeddingNative(dimension=2)])

    """
    if method is None:
        method = StochasticProximityEmbeddingNative()
    if ensembles is not None:
        if not hasattr(ensembles, "__iter__"):
            ensembles = [ensembles]

        ensembles_list = ensembles
        if not hasattr(ensembles[0], "__iter__"):
            ensembles_list = [ensembles]

        # Calculate merged ensembles and transfer to memory
        merged_ensembles = []
        for ensembles in ensembles_list:
            # Transfer ensembles to memory
            for ensemble in ensembles:
                ensemble.transfer_to_memory()
            merged_ensembles.append(merge_universes(ensembles))

    methods = method
    if not hasattr(method, "__iter__"):
        methods = [method]

    # Check whether any of the methods can make use of a distance matrix
    any_method_accept_distance_matrix = np.any(
        [_method.accepts_distance_matrix for _method in methods]
    )

    # If distance matrices are provided, check that it matches the number
    # of ensembles
    if distance_matrix:
        if not hasattr(distance_matrix, "__iter__"):
            distance_matrix = [distance_matrix]
        if ensembles is not None and len(distance_matrix) != len(
            merged_ensembles
        ):
            raise ValueError(
                "Dimensions of provided list of distance matrices "
                "does not match that of provided list of "
                "ensembles: {0} vs {1}".format(
                    len(distance_matrix), len(merged_ensembles)
                )
            )

    else:
        # Calculate distance matrices for all merged ensembles - if not provided
        if any_method_accept_distance_matrix:
            distance_matrix = []
            for merged_ensemble in merged_ensembles:
                distance_matrix.append(
                    get_distance_matrix(
                        merged_ensemble, select=select, **kwargs
                    )
                )

    args = []
    for method in methods:
        if method.accepts_distance_matrix:
            args += [(d,) for d in distance_matrix]
        else:
            for merged_ensemble in merged_ensembles:
                coordinates = merged_ensemble.trajectory.timeseries(
                    order="fac"
                )

                # Flatten coordinate matrix into n_frame x n_coordinates
                coordinates = np.reshape(
                    coordinates, (coordinates.shape[0], -1)
                )

                args.append((coordinates,))

    # Execute dimensionality reduction procedure
    pc = ParallelCalculation(ncores, methods, args)

    # Run parallel calculation
    results = pc.run()

    # Keep track of which sample belongs to which ensembles
    details = {}
    if ensembles is not None:
        ensemble_assignment = []
        for i, ensemble in enumerate(ensembles):
            ensemble_assignment += [i + 1] * len(ensemble.trajectory)
        ensemble_assignment = np.array(ensemble_assignment)
        details["ensemble_membership"] = ensemble_assignment

    coordinates = []
    for result in results:
        coordinates.append(result[1][0])
        # details.append(result[1][1])

    if allow_collapsed_result and len(coordinates) == 1:
        coordinates = coordinates[0]
        # details = details[0]

    return coordinates, details



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""DCD trajectory I/O  --- :mod:`MDAnalysis.coordinates.DCD`
============================================================

Classes to read and write DCD binary trajectories, the format used by
CHARMM, NAMD, and also LAMMPS. Trajectories can be read regardless of
system-endianness as this is auto-detected.

Generally, DCD trajectories produced by any code can be read (with the
:class:`DCDReader`) although there can be issues with the unitcell (simulation
box) representation (see :attr:`DCDReader.dimensions`). DCDs can also be
written but the :class:`DCDWriter` follows recent NAMD/VMD convention for the
unitcell while still writing AKMA time. Reading and writing these trajectories
within MDAnalysis will work seamlessly but if you process those trajectories
with other tools you might need to watch out that time and unitcell dimensions
are correctly interpreted.


See Also
--------
:mod:`MDAnalysis.coordinates.LAMMPS`
  module provides a more flexible DCD reader/writer.


.. _Issue 187:
   https://github.com/MDAnalysis/mdanalysis/issues/187


Classes
-------

.. autoclass:: DCDReader
   :inherited-members:
.. autoclass:: DCDWriter
   :inherited-members:

"""
import os
import errno
import numpy as np
import struct
import types
import warnings

from .. import units as mdaunits  # use mdaunits instead of units to avoid a clash
from . import base, core
from ..lib.formats.libdcd import DCDFile
from ..lib.mdamath import triclinic_box
from ..lib.util import store_init_arguments


class DCDReader(base.ReaderBase):
    """Reader for the DCD format.

    DCD is used by NAMD, CHARMM and LAMMPS as the default trajectory format.
    The DCD file format is not well defined. In particular, NAMD and CHARMM use
    it differently. Currently, MDAnalysis tries to guess the correct **format
    for the unitcell representation** but it can be wrong. **Check the unitcell
    dimensions**, especially for triclinic unitcells (see `Issue 187`_). DCD
    trajectories produced by CHARMM and NAMD( >2.5) record time in AKMA units.
    If other units have been recorded (e.g., ps) then employ the configurable
    :class:`MDAnalysis.coordinates.LAMMPS.DCDReader` and set the time unit as a
    optional argument. You can find a list of units used in the DCD formats on
    the MDAnalysis `wiki`_.

    MDAnalysis always uses (*A*, *B*, *C*, α, β, γ) to represent the unit
    cell. Lengths *A*, *B*, *C* are in the MDAnalysis length unit (Å), and
    angles α, β, γ are in degrees.

    The ordering of the angles in the unitcell is the same as in recent
    versions of VMD's DCDplugin_ (2013), namely the `X-PLOR DCD format`_: The
    original unitcell is read as ``[A, gamma, B, beta, alpha, C]`` from the DCD
    file. If any of these values are < 0 or if any of the angles are > 180º
    then it is assumed it is a new-style CHARMM unitcell (possibly ≥22, at
    least since c36b2) in which symmetric box vectors were recorded. If all
    angles are numbers between -1 and +1, then they are treated as cosines of
    the angle (following the modern NAMD > 2.5/VMD convention).

    Ultimately, all unitcell representations are converted to the MDAnalysis
    standard unit cell representation in :attr:`DCDReader.dimensions`.


    .. deprecated:: 2.4.0
        DCDReader currently makes independent timesteps
        by copying the :class:`Timestep` associated with the reader.
        Other readers update the :class:`Timestep` inplace meaning all
        references to the :class:`Timestep` contain the same data. The unique
        independent :class:`Timestep` behaviour of the DCDReader is deprecated
        will be changed in 3.0 to be the same as other readers

    .. warning::
        The DCD format is not well defined. Check your unit cell
        dimensions carefully, especially when using triclinic boxes.
        Different software packages implement different conventions and
        MDAnalysis is currently implementing the newer NAMD/VMD convention
        and tries to guess the new CHARMM one. Old CHARMM trajectories might
        give wrong unitcell values. For more details see `Issue 187`_.

    .. _`X-PLOR DCD format`: http://www.ks.uiuc.edu/Research/vmd/plugins/molfile/dcdplugin.html
    .. _Issue 187: https://github.com/MDAnalysis/mdanalysis/issues/187
    .. _DCDplugin: http://www.ks.uiuc.edu/Research/vmd/plugins/doxygen/dcdplugin_8c-source.html#l00947
    .. _wiki: https://github.com/MDAnalysis/mdanalysis/wiki/FileFormats#dcd

    """
    format = 'DCD'
    flavor = 'CHARMM'
    units = {'time': 'AKMA', 'length': 'Angstrom'}

    @store_init_arguments
    def __init__(self, filename, convert_units=True, dt=None, **kwargs):
        """
        Parameters
        ----------
        filename : str
            trajectory filename
        convert_units : bool (optional)
            convert units to MDAnalysis units
        dt : float (optional)
            overwrite time delta stored in DCD
        **kwargs : dict
            General reader arguments.


        .. versionchanged:: 0.17.0
           Changed to use libdcd.pyx library and removed the correl function
        """
        super(DCDReader, self).__init__(
            filename, convert_units=convert_units, **kwargs)
        self._file = DCDFile(self.filename)
        self.n_atoms = self._file.header['natoms']

        delta = mdaunits.convert(self._file.header['delta'],
                                 self.units['time'], 'ps')
        if dt is None:
            dt = delta * self._file.header['nsavc']
        self.skip_timestep = self._file.header['nsavc']

        self._ts_kwargs['dt'] = dt
        self.ts = self._Timestep(self.n_atoms, **self._ts_kwargs)
        frame = self._file.read()
        # reset trajectory
        if self._file.n_frames > 1:
            self._file.seek(1)
        else:
            self._file.seek(0)
        self._frame = 0
        self.ts = self._frame_to_ts(frame, self.ts)
        # these should only be initialized once
        self.ts.dt = dt
        warnings.warn("DCDReader currently makes independent timesteps"
                      " by copying self.ts while other readers update"
                      " self.ts inplace. This behavior will be changed in"
                      " 3.0 to be the same as other readers. Read more at"
                      " https://github.com/MDAnalysis/mdanalysis/issues/3889"
                      " to learn if this change in behavior might affect you.",
                       category=DeprecationWarning)

    @staticmethod
    def parse_n_atoms(filename, **kwargs):
        with DCDFile(filename) as f:
            n_atoms = f.header['natoms']
        return n_atoms

    def close(self):
        """close reader"""
        self._file.close()

    @property
    def n_frames(self):
        """number of frames in trajectory"""
        return len(self._file)

    def _reopen(self):
        """reopen trajectory"""
        self.ts.frame = 0
        self._frame = -1
        self._file.close()
        self._file.open('r')

    def _read_frame(self, i):
        """read frame i"""
        self._frame = i - 1
        self._file.seek(i)
        return self._read_next_timestep()

    def _read_next_timestep(self, ts=None):
        """copy next frame into timestep"""
        if self._frame == self.n_frames - 1:
            raise IOError('trying to go over trajectory limit')
        if ts is None:
            #TODO remove copying the ts in 3.0 
            ts = self.ts.copy()
        frame = self._file.read()
        self._frame += 1
        self._frame_to_ts(frame, ts)
        self.ts = ts
        return ts

    def Writer(self, filename, n_atoms=None, **kwargs):
        """Return writer for trajectory format"""
        if n_atoms is None:
            n_atoms = self.n_atoms
        return DCDWriter(
            filename,
            n_atoms=n_atoms,
            dt=self.ts.dt,
            convert_units=self.convert_units,
            **kwargs)

    def _frame_to_ts(self, frame, ts):
        """convert a dcd-frame to a :class:`TimeStep`"""
        ts.frame = self._frame
        ts.time = (ts.frame + self._file.header['istart']/self._file.header['nsavc']) * self.ts.dt
        ts.data['step'] = self._file.tell()

        # The original unitcell is read as ``[A, gamma, B, beta, alpha, C]``
        _ts_order = [0, 2, 5, 4, 3, 1]
        uc = np.take(frame.unitcell, _ts_order)

        pi_2 = np.pi / 2
        if (-1.0 <= uc[3] <= 1.0) and (-1.0 <= uc[4] <= 1.0) and (
                -1.0 <= uc[5] <= 1.0):
            # This file was generated by Charmm, or by NAMD > 2.5, with the
            # angle cosines of the periodic cell angles written to the DCD
            # file. This formulation improves rounding behavior for orthogonal
            # cells so that the angles end up at precisely 90 degrees, unlike
            # acos(). (changed in MDAnalysis 0.9.0 to have NAMD ordering of the
            # angles; see Issue 187) */
            uc[3] = 90.0 - np.arcsin(uc[3]) * 90.0 / pi_2
            uc[4] = 90.0 - np.arcsin(uc[4]) * 90.0 / pi_2
            uc[5] = 90.0 - np.arcsin(uc[5]) * 90.0 / pi_2
        # heuristic sanity check: uc = A,B,C,alpha,beta,gamma
        elif np.any(uc < 0.) or np.any(uc[3:] > 180.):
            # might be new CHARMM: box matrix vectors
            H = frame.unitcell.copy()
            e1, e2, e3 = H[[0, 1, 3]], H[[1, 2, 4]], H[[3, 4, 5]]
            uc = triclinic_box(e1, e2, e3)
        else:
            # This file was likely generated by NAMD 2.5 and the periodic cell
            # angles are specified in degrees rather than angle cosines.
            pass

        ts.dimensions = uc
        ts.positions = frame.xyz

        if self.convert_units:
            if ts.dimensions is not None:
                self.convert_pos_from_native(ts.dimensions[:3])
            self.convert_pos_from_native(ts.positions)

        return ts

    @property
    def dimensions(self):
        """unitcell dimensions (*A*, *B*, *C*, *alpha*, *beta*, *gamma*)
        """
        return self.ts.dimensions

    @property
    def dt(self):
        """timestep between frames"""
        return self.ts.dt

    def timeseries(self,
                   asel=None,
                   atomgroup=None,
                   start=None,
                   stop=None,
                   step=None,
                   order='afc'):
        """Return a subset of coordinate data for an AtomGroup

        Parameters
        ----------
        asel : :class:`~MDAnalysis.core.groups.AtomGroup`
            The :class:`~MDAnalysis.core.groups.AtomGroup` to read the
            coordinates from. Defaults to None, in which case the full set of
            coordinate data is returned.
        
            .. deprecated:: 2.7.0
               asel argument will be renamed to atomgroup in 3.0.0

        atomgroup: AtomGroup (optional)
            Same as `asel`, will replace `asel` in 3.0.0
        start : int (optional)
            Begin reading the trajectory at frame index `start` (where 0 is the
            index of the first frame in the trajectory); the default ``None``
            starts at the beginning.
        stop : int (optional)
            End reading the trajectory at frame index `stop`-1, i.e, `stop` is
            excluded. The trajectory is read to the end with the default
            ``None``.
        step : int (optional)
            Step size for reading; the default ``None`` is equivalent to 1 and
            means to read every frame.
        order : str (optional)
            the order/shape of the return data array, corresponding
            to (a)tom, (f)rame, (c)oordinates all six combinations
            of 'a', 'f', 'c' are allowed ie "fac" - return array
            where the shape is (frame, number of atoms,
            coordinates)


        .. versionchanged:: 1.0.0
           `skip` and `format` keywords have been removed.
        .. versionchanged:: 2.4.0
            ValueError now raised instead of NoDataError for empty input
            AtomGroup
        """
        if asel is not None:
            warnings.warn(
                "asel argument to timeseries will be renamed to"
                "'atomgroup' in 3.0, see #3911",
                category=DeprecationWarning)
            if atomgroup:
                raise ValueError("Cannot provide both asel and atomgroup kwargs")
            atomgroup = asel

        start, stop, step = self.check_slice_indices(start, stop, step)

        if atomgroup is not None:
            if len(atomgroup) == 0:
                raise ValueError(
                    "Timeseries requires at least one atom to analyze")
            atom_numbers = list(atomgroup.indices)
        else:
            atom_numbers = list(range(self.n_atoms))

        frames = self._file.readframes(
            start, stop, step, order=order, indices=atom_numbers)
        return frames.xyz


class DCDWriter(base.WriterBase):
    """DCD Writer class

    The writer follows recent NAMD/VMD convention for the unitcell (box lengths
    in Å and angle-cosines, ``[A, cos(gamma), B, cos(beta), cos(alpha), C]``)
    and writes positions in Å and time in AKMA time units.

    .. warning::
    
       Multiple conventions exist for how unit cells are written to DCD
       files. Until 2.10.0 MDAnalysis followed the old X-PLOR/CHARMM/NAMD
       (≤2.5) convention of storing ``[A, gamma, B, beta, alpha, C]`` while
       wrongly stating that the new NAMD/VMD convention would be followed. In
       2.10.0, MDAnalysis switched to following the documented behavior and now
       stores the box with angle cosines, as described above. The MDAnalysis
       :class:`DCDReader` can correctly read either format but *if you have
       code that relies on a specific format for the box information in the raw
       DCD file, please check your results.*

       Furthermore, modern versions of CHARMM store box vectors and not box
       length/angles. When reading MDAnalysis-generated DCD files *in CHARMM*,
       carefully check your results.

    .. note::
        When writing out timesteps without ``dimensions`` (i.e. set ``None``)
        the :class:`DCDWriter` will write out a zeroed unitcell (i.e.
        ``[0, 0, 0, 0, 0, 0]``). As this behaviour is poorly defined, it may
        not match the expectations of other software.

    .. versionchanged:: 2.10.0    
       Up to 2.10.0 the :class:`DCDWriter` wrote a unit cell following old NAMD
       (≤2.5) convention, even though the docs stated that the new NAMD
       convention was being used. Now the modern NAMD > 2.5 format is
       written. See `Issue #5069`_ for details.


    .. _`Issue #5069`: https://github.com/MDAnalysis/mdanalysis/issues/5069

    """
    format = 'DCD'
    multiframe = True
    flavor = 'NAMD'
    units = {'time': 'AKMA', 'length': 'Angstrom'}

    def __init__(self,
                 filename,
                 n_atoms,
                 convert_units=True,
                 step=1,
                 dt=1,
                 remarks='',
                 nsavc=1,
                 istart=0,
                 **kwargs):
        """Parameters
        ----------
        filename : str
            filename of trajectory
        n_atoms : int
            number of atoms to be written
        convert_units : bool (optional)
            convert from MDAnalysis units to format specific units
        step : int (optional)
            number of steps between frames to be written
        dt : float (optional)
            time between two frames. If ``None`` guess from first written
            :class:`TimeStep`
        remarks : str (optional)
            remarks to be stored in DCD. Shouldn't be more then 240 characters
        nsavc : int (optional)
            DCD files can also store the number of integrator time steps that
            correspond to the interval between two frames as nsavc (i.e., every
            how many MD steps is a frame saved to the DCD). By default, this
            number is just set to one and this should be sufficient for almost
            all cases but if required, `nsavc` can be changed.
        istart : int (optional)
            starting frame number in integrator timesteps. CHARMM defaults to
            `nsavc`, i.e., start at frame number 1 = `istart` / `nsavc`. The value
            ``None`` will set `istart` to `nsavc` (the CHARMM default).
            The MDAnalysis default is 0 so that the frame number and time of the first
            frame is 0.
        **kwargs : dict
            General writer arguments

        """
        self.filename = filename
        self._convert_units = convert_units
        if n_atoms is None:
            raise ValueError("n_atoms argument is required")
        self.n_atoms = n_atoms
        self._file = DCDFile(self.filename, 'w')
        self.step = step
        self.dt = dt
        dt = mdaunits.convert(dt, 'ps', self.units['time'])
        delta = float(dt) / nsavc
        istart = istart if istart is not None else nsavc
        self._file.write_header(
            remarks=remarks,
            natoms=self.n_atoms,
            nsavc=nsavc,
            delta=delta,
            is_periodic=1,
            istart=istart)

    def _write_next_frame(self, ag):
        """Write information associated with ``obj`` at current frame into trajectory

        Parameters
        ----------
        ag : AtomGroup or Universe

        See Also
        --------
        :meth:`DCDWriter.write`  takes a more general input


        .. versionchanged:: 1.0.0
           Added ability to pass AtomGroup or Universe.
           Renamed from `write_next_timestep` to `_write_next_frame`.
        .. versionchanged:: 2.0.0
           Deprecated support for Timestep argument has now been removed.
           Use AtomGroup or Universe as an input instead.
        """
        try:
            ts = ag.ts
        except AttributeError:
            try:
                # Universe?
                ts = ag.trajectory.ts
            except AttributeError:
                errmsg = "Input obj is neither an AtomGroup or Universe"
                raise TypeError(errmsg) from None
        xyz = ts.positions.copy()
        try:
            dimensions = ts.dimensions.copy()
        except AttributeError:
            wmsg = ('No dimensions set for current frame, zeroed unitcell '
                    'will be written')
            warnings.warn(wmsg)
            dimensions = np.zeros(6)

        if self._convert_units:
            xyz = self.convert_pos_to_native(xyz, inplace=True)
            dimensions = self.convert_dimensions_to_unitcell(ts, inplace=True)

        # Convert angles to cosines following NAMD/VMD convention
        # The DCD unitcell is written as [A, cos(gamma), B, cos(beta), cos(alpha), C]
        _ts_order = [0, 5, 1, 4, 3, 2]
        box = np.take(dimensions, _ts_order)
        
        # Convert angles (indices 1, 3, 4) from degrees to the special cosine format
        # used by NAMD/VMD: cos(angle) = sin(90 - angle); see Issue 5069
        box[[1, 3, 4]] = np.sin(np.deg2rad(90.0 - box[[1, 3, 4]]))

        self._file.write(xyz=xyz, box=box)

    def close(self):
        """close trajectory"""
        self._file.close()

    def __del__(self):
        self.close()



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

import warnings
import pytest
import numpy as np
from numpy.testing import assert_equal

import MDAnalysis as mda
from MDAnalysisTests.topology.base import ParserBase
from MDAnalysisTests.datafiles import mol2_molecule, PDB_helix, SDF_molecule

# TODO: remove these shims when RDKit
# has a release supporting NumPy 2
Chem = pytest.importorskip("rdkit.Chem")
AllChem = pytest.importorskip("rdkit.Chem.AllChem")


class RDKitParserBase(ParserBase):
    parser = mda.converters.RDKitParser.RDKitParser
    expected_attrs = [
        "ids",
        "names",
        "elements",
        "masses",
        "aromaticities",
        "resids",
        "resnums",
        "chiralities",
        "segids",
        "bonds",
    ]

    expected_n_atoms = 0
    expected_n_residues = 1
    expected_n_segments = 1
    expected_n_bonds = 0

    @pytest.fixture()
    def top(self, filename):
        with self.parser(filename) as p:
            yield p.parse()

    def test_creates_universe(self, filename):
        u = mda.Universe(filename, format="RDKIT")
        assert isinstance(u, mda.Universe)

    def test_bonds_total_counts(self, top):
        assert len(top.bonds.values) == self.expected_n_bonds

    def test_guessed_attributes(self, filename):
        u = mda.Universe(filename, format="RDKIT")
        u_guessed_attrs = [a.attrname for a in u._topology.guessed_attributes]
        for attr in self.guessed_attrs:
            assert hasattr(u.atoms, attr)
            assert attr in u_guessed_attrs


class TestRDKitParserMOL2(RDKitParserBase):
    ref_filename = mol2_molecule

    expected_attrs = RDKitParserBase.expected_attrs + ["charges", "types"]

    expected_n_atoms = 49
    expected_n_residues = 1
    expected_n_segments = 1
    expected_n_bonds = 51

    @pytest.fixture
    def filename(self):
        return Chem.MolFromMol2File(self.ref_filename, removeHs=False)

    def _create_mol_gasteiger_charges(self):
        mol = Chem.MolFromMol2File(self.ref_filename, removeHs=False)
        AllChem.ComputeGasteigerCharges(mol)
        return mol

    def _remove_tripos_charges(self, mol):
        for atom in mol.GetAtoms():
            atom.ClearProp("_TriposPartialCharge")

    @pytest.fixture
    def top_gas_tripos(self):
        mol = self._create_mol_gasteiger_charges()
        return self.parser(mol).parse()

    @pytest.fixture
    def filename_gasteiger(self):
        mol = self._create_mol_gasteiger_charges()
        self._remove_tripos_charges(mol)
        return mol

    @pytest.fixture
    def top_gasteiger(self):
        mol = self._create_mol_gasteiger_charges()
        self._remove_tripos_charges(mol)
        return self.parser(mol).parse()

    def test_bond_orders(self, top, filename):
        expected = [bond.GetBondTypeAsDouble() for bond in filename.GetBonds()]
        assert top.bonds.order == expected

    def test_multiple_charge_priority(
        self, top_gas_tripos, filename_gasteiger
    ):
        expected = np.array(
            [
                a.GetDoubleProp("_GasteigerCharge")
                for a in filename_gasteiger.GetAtoms()
            ],
            dtype=np.float32,
        )
        assert_equal(expected, top_gas_tripos.charges.values)

    def test_multiple_charge_props_warning(self):
        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            mol = self._create_mol_gasteiger_charges()
            # Trigger a warning.
            top = self.parser(mol).parse()
            # Verify the warning
            assert len(w) == 1
            assert "_GasteigerCharge and _TriposPartialCharge" in str(
                w[-1].message
            )

    def test_gasteiger_charges(self, top_gasteiger, filename_gasteiger):
        expected = np.array(
            [
                a.GetDoubleProp("_GasteigerCharge")
                for a in filename_gasteiger.GetAtoms()
            ],
            dtype=np.float32,
        )
        assert_equal(expected, top_gasteiger.charges.values)

    def test_tripos_charges(self, top, filename):
        expected = np.array(
            [
                a.GetDoubleProp("_TriposPartialCharge")
                for a in filename.GetAtoms()
            ],
            dtype=np.float32,
        )
        assert_equal(expected, top.charges.values)

    def test_aromaticity(self, top, filename):
        expected = np.array(
            [atom.GetIsAromatic() for atom in filename.GetAtoms()]
        )
        assert_equal(expected, top.aromaticities.values)

    def test_guessed_types(self, filename):
        u = mda.Universe(filename, format="RDKIT")
        assert_equal(
            u.atoms.types[:7],
            ["N.am", "S.o2", "N.am", "N.am", "O.2", "O.2", "C.3"],
        )


class TestRDKitParserPDB(RDKitParserBase):
    ref_filename = PDB_helix

    expected_attrs = RDKitParserBase.expected_attrs + [
        "resnames",
        "altLocs",
        "chainIDs",
        "occupancies",
        "icodes",
        "tempfactors",
    ]

    expected_n_atoms = 137
    expected_n_residues = 13
    expected_n_segments = 1
    expected_n_bonds = 137

    @pytest.fixture
    def filename(self):
        return Chem.MolFromPDBFile(self.ref_filename, removeHs=False)

    def test_partial_residueinfo_raise_error(self, filename):
        mol = Chem.RemoveHs(filename)
        mh = Chem.AddHs(mol)
        with pytest.raises(
            ValueError, match="ResidueInfo is only partially available"
        ):
            mda.Universe(mh)
        mh = Chem.AddHs(mol, addResidueInfo=True)
        mda.Universe(mh)

    def test_guessed_types(self, filename):
        u = mda.Universe(filename, format="RDKIT")
        assert_equal(u.atoms.types[:7], ["N", "H", "C", "H", "C", "H", "H"])


class TestRDKitParserSMILES(RDKitParserBase):
    ref_filename = "CN1C=NC2=C1C(=O)N(C(=O)N2C)C"

    expected_n_atoms = 24
    expected_n_residues = 1
    expected_n_segments = 1
    expected_n_bonds = 25

    @pytest.fixture
    def filename(self):
        mol = Chem.MolFromSmiles(self.ref_filename)
        mol = Chem.AddHs(mol)
        return mol


class TestRDKitParserSDF(RDKitParserBase):
    ref_filename = SDF_molecule

    expected_n_atoms = 49
    expected_n_residues = 1
    expected_n_segments = 1
    expected_n_bonds = 49

    @pytest.fixture
    def filename(self):
        return Chem.SDMolSupplier(SDF_molecule, removeHs=False)[0]

    def test_bond_orders(self, top, filename):
        expected = [bond.GetBondTypeAsDouble() for bond in filename.GetBonds()]
        assert top.bonds.order == expected

    def test_guessed_types(self, filename):
        u = mda.Universe(filename, format="RDKIT")
        assert_equal(u.atoms.types[:7], ["CA", "C", "C", "C", "C", "C", "O"])



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
PSF topology parser
===================

Reads a CHARMM/NAMD/XPLOR PSF_ file to build the system. The topology will
contain atom IDs, segids, residue IDs, residue names, atom names, atom types,
charges and masses. Bonds, angles, dihedrals and impropers are also read from
the file.

It reads both standard and extended ("EXT") PSF formats and can also parse NAMD
space-separated "PSF" file variants.

.. _PSF: http://www.charmm.org/documentation/c35b1/struct.html

Classes
-------

.. autoclass:: PSFParser
   :members:
   :inherited-members:

"""
import logging
import functools
from math import ceil
import numpy as np

from ..lib.util import openany, atoi

from .base import TopologyReaderBase, squash_by, change_squash
from ..core.topologyattrs import (
    Atomids,
    Atomnames,
    Atomtypes,
    Masses,
    Charges,
    Resids,
    Resnums,
    Resnames,
    Segids,
    Bonds,
    Angles,
    Dihedrals,
    Impropers,
)
from ..core.topology import Topology

logger = logging.getLogger("MDAnalysis.topology.PSF")


class PSFParser(TopologyReaderBase):
    """Read topology information from a CHARMM/NAMD/XPLOR PSF_ file.

    Creates a Topology with the following Attributes:
    - ids
    - names
    - types
    - masses
    - charges
    - resids
    - resnames
    - segids
    - bonds
    - angles
    - dihedrals
    - impropers

    .. _PSF: http://www.charmm.org/documentation/c35b1/struct.html


    .. versionchanged:: 2.8.0
       PSFParser now reads string resids and converts them to integers.
    """

    format = "PSF"

    def parse(self, **kwargs):
        """Parse PSF file into Topology

        Returns
        -------
        MDAnalysis *Topology* object
        """
        # Open and check psf validity
        with openany(self.filename) as psffile:
            header = next(psffile)
            if not header.startswith("PSF"):
                err = "{0} is not valid PSF file (header = {1})" "".format(
                    self.filename, header
                )
                logger.error(err)
                raise ValueError(err)
            header_flags = header[3:].split()

            if "NAMD" in header_flags:
                self._format = "NAMD"  # NAMD/VMD
            elif "EXT" in header_flags:
                self._format = "EXTENDED"  # CHARMM
            else:
                self._format = "STANDARD"  # CHARMM

            next(psffile)
            title = next(psffile).split()
            if not (title[1] == "!NTITLE"):
                err = "{0} is not a valid PSF file".format(self.filename)
                logger.error(err)
                raise ValueError(err)
            # psfremarks = [psffile.next() for i in range(int(title[0]))]
            for _ in range(int(title[0])):
                next(psffile)
            logger.debug(
                "PSF file {0}: format {1}"
                "".format(self.filename, self._format)
            )

            # Atoms first and mandatory
            top = self._parse_sec(psffile, ("NATOM", 1, 1, self._parseatoms))
            # Then possibly other sections
            sections = (
                # ("atoms", ("NATOM", 1, 1, self._parseatoms)),
                (Bonds, ("NBOND", 2, 4, self._parsesection)),
                (Angles, ("NTHETA", 3, 3, self._parsesection)),
                (Dihedrals, ("NPHI", 4, 2, self._parsesection)),
                (Impropers, ("NIMPHI", 4, 2, self._parsesection)),
                # ("donors", ("NDON", 2, 4, self._parsesection)),
                # ("acceptors", ("NACC", 2, 4, self._parsesection))
            )

            try:
                for attr, info in sections:
                    next(psffile)
                    top.add_TopologyAttr(attr(self._parse_sec(psffile, info)))
            except StopIteration:
                # Reached the end of the file before we expected
                for attr in (Bonds, Angles, Dihedrals, Impropers):
                    if not hasattr(top, attr.attrname):
                        top.add_TopologyAttr(attr([]))

        return top

    def _parse_sec(self, psffile, section_info):
        """Parse a single section of the PSF

        Returns
        -------
        A list of Attributes from this section
        """
        desc, atoms_per, per_line, parsefunc = section_info
        header = next(psffile)
        while header.strip() == "":
            header = next(psffile)
        header = header.split()
        # Get the number
        num = float(header[0])
        sect_type = header[1].strip("!:")
        # Make sure the section type matches the desc
        if not sect_type == desc:
            err = "Expected section {0} but found {1}" "".format(
                desc, sect_type
            )
            logger.error(err)
            raise ValueError(err)
        # Now figure out how many lines to read
        numlines = int(ceil(num / per_line))

        psffile_next = functools.partial(next, psffile)
        return parsefunc(psffile_next, atoms_per, numlines)

    def _parseatoms(self, lines, atoms_per, numlines):
        """Parses atom section in a Charmm PSF file.

        Normal (standard) and extended (EXT) PSF format are
        supported. CHEQ is supported in the sense that CHEQ data is simply
        ignored.


        CHARMM Format from ``source/psffres.src``:

        CHEQ::
          II,LSEGID,LRESID,LRES,TYPE(I),IAC(I),CG(I),AMASS(I),IMOVE(I),ECH(I),EHA(I)

          standard format:
            (I8,1X,A4,1X,A4,1X,A4,1X,A4,1X,I4,1X,2G14.6,I8,2G14.6)
            (I8,1X,A4,1X,A4,1X,A4,1X,A4,1X,A4,1X,2G14.6,I8,2G14.6)  XPLOR
          expanded format EXT:
            (I10,1X,A8,1X,A8,1X,A8,1X,A8,1X,I4,1X,2G14.6,I8,2G14.6)
            (I10,1X,A8,1X,A8,1X,A8,1X,A8,1X,A4,1X,2G14.6,I8,2G14.6) XPLOR

        no CHEQ::
          II,LSEGID,LRESID,LRES,TYPE(I),IAC(I),CG(I),AMASS(I),IMOVE(I)

         standard format:
            (I8,1X,A4,1X,A4,1X,A4,1X,A4,1X,I4,1X,2G14.6,I8)
            (I8,1X,A4,1X,A4,1X,A4,1X,A4,1X,A4,1X,2G14.6,I8)  XPLOR
          expanded format EXT:
            (I10,1X,A8,1X,A8,1X,A8,1X,A8,1X,I4,1X,2G14.6,I8)
            (I10,1X,A8,1X,A8,1X,A8,1X,A8,1X,A4,1X,2G14.6,I8) XPLOR

        NAMD PSF

        space separated, see release notes for VMD 1.9.1, psfplugin at
        http://www.ks.uiuc.edu/Research/vmd/current/devel.html :

        psfplugin: Added more logic to the PSF plugin to determine cases where the
        CHARMM "EXTended" PSF format cannot accomodate long atom types, and we add
        a "NAMD" keyword to the PSF file flags line at the top of the file. Upon
        reading, if we detect the "NAMD" flag there, we know that it is possible
        to parse the file correctly using a simple space-delimited scanf() format
        string, and we use that strategy rather than holding to the inflexible
        column-based fields that are a necessity for compatibility with CHARMM,
        CNS, X-PLOR, and other formats. NAMD and the psfgen plugin already assume
        this sort of space-delimited formatting, but that's because they aren't
        expected to parse the PSF variants associated with the other programs. For
        the VMD PSF plugin, having the "NAMD" tag in the flags line makes it
        absolutely clear that we're dealing with a NAMD-specific file so we can
        take the same approach.

        """
        # how to partition the line into the individual atom components
        atom_parsers = {
            "STANDARD": lambda l: (
                l[:8],
                l[9:13].strip() or "SYSTEM",
                l[14:18],
                l[19:23].strip(),
                l[24:28].strip(),
                l[29:33].strip(),
                l[34:48],
                l[48:62],
            ),
            # l[62:70], l[70:84], l[84:98] ignore IMOVE, ECH and EHA,
            "EXTENDED": lambda l: (
                l[:10],
                l[11:19].strip() or "SYSTEM",
                l[20:28],
                l[29:37].strip(),
                l[38:46].strip(),
                l[47:51].strip(),
                l[52:66],
                l[66:70],
            ),
            # l[70:78],  l[78:84], l[84:98] ignore IMOVE, ECH and EHA,
            "NAMD": lambda l: l.split()[:8],
        }
        atom_parser = atom_parsers[self._format]
        # once partitioned, assigned each component the correct type
        set_type = lambda x: (
            int(x[0]) - 1,
            x[1] or "SYSTEM",
            atoi(x[2]),
            x[3],
            x[4],
            x[5],
            float(x[6]),
            float(x[7]),
        )

        # Oli: I don't think that this is the correct OUTPUT format:
        #   psf_atom_format = "   %5d %4s %4d %4s %-4s %-4s %10.6f      %7.4f%s\n"
        # It should be rather something like:
        #   psf_ATOM_format = '%(iatom)8d %(segid)4s %(resid)-4d %(resname)4s '+\
        #                     '%(name)-4s %(type)4s %(charge)-14.6f%(mass)-14.4f%(imove)8d\n'

        # source/psfres.src (CHEQ and now can be used for CHEQ EXTended), see comments above
        #   II,LSEGID,LRESID,LRES,TYPE(I),IAC(I),CG(I),AMASS(I),IMOVE(I),ECH(I),EHA(I)
        #  (I8,1X,A4, 1X,A4,  1X,A4,  1X,A4,  1X,I4,  1X,2G14.6,     I8,   2G14.6)
        #   0:8   9:13   14:18   19:23   24:28   29:33   34:48 48:62 62:70 70:84 84:98

        # Allocate arrays
        atomids = np.zeros(numlines, dtype=np.int32)
        segids = np.zeros(numlines, dtype=object)
        resids = np.zeros(numlines, dtype=np.int32)
        resnames = np.zeros(numlines, dtype=object)
        atomnames = np.zeros(numlines, dtype=object)
        atomtypes = np.zeros(numlines, dtype=object)
        charges = np.zeros(numlines, dtype=np.float32)
        masses = np.zeros(numlines, dtype=np.float64)

        for i in range(numlines):
            try:
                line = lines()
            except StopIteration:
                err = f"{self.filename} is not valid PSF file"
                logger.error(err)
                raise ValueError(err) from None
            try:
                vals = set_type(atom_parser(line))
            except ValueError:
                # last ditch attempt: this *might* be a NAMD/VMD
                # space-separated "PSF" file from VMD version < 1.9.1
                atom_parser = atom_parsers["NAMD"]
                vals = set_type(atom_parser(line))
                logger.warning(
                    "Guessing that this is actually a"
                    " NAMD-type PSF file..."
                    " continuing with fingers crossed!"
                )
                logger.debug(
                    "First NAMD-type line: {0}: {1}"
                    "".format(i, line.rstrip())
                )

            atomids[i] = vals[0]
            segids[i] = vals[1]
            resids[i] = vals[2]
            resnames[i] = vals[3]
            atomnames[i] = vals[4]
            atomtypes[i] = vals[5]
            charges[i] = vals[6]
            masses[i] = vals[7]

        # Atom
        atomids = Atomids(atomids)
        atomnames = Atomnames(atomnames)
        atomtypes = Atomtypes(atomtypes)
        charges = Charges(charges)
        masses = Masses(masses)

        # Residue
        # resids, resnames
        residx, (new_resids, new_resnames, perres_segids) = change_squash(
            (resids, resnames, segids), (resids, resnames, segids)
        )
        # transform from atom:Rid to atom:Rix
        residueids = Resids(new_resids)
        residuenums = Resnums(new_resids.copy())
        residuenames = Resnames(new_resnames)

        # Segment
        segidx, perseg_segids = squash_by(perres_segids)[:2]
        segids = Segids(perseg_segids)

        top = Topology(
            len(atomids),
            len(new_resids),
            len(segids),
            attrs=[
                atomids,
                atomnames,
                atomtypes,
                charges,
                masses,
                residueids,
                residuenums,
                residuenames,
                segids,
            ],
            atom_resindex=residx,
            residue_segindex=segidx,
        )

        return top

    def _parsesection(self, lines, atoms_per, numlines):
        section = []

        for i in range(numlines):
            # Subtract 1 from each number to ensure zero-indexing for the atoms
            fields = np.int64(lines().split()) - 1
            for j in range(0, len(fields), atoms_per):
                section.append(tuple(fields[j : j + atoms_per]))
        return section



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
TNG trajectory files --- :mod:`MDAnalysis.coordinates.TNG`
==========================================================


The TNG format :footcite:p:`Lundborg2014` is a format used in `GROMACS`_ for
storage of trajectory and topology information. The TNG format allows a wide
range of compression algorithms and unlike the compressed XTC format can also
store velocities and forces in addition to positions.

The classes in this module are based on the `pytng`_ package for reading TNG
files. The reader is directed to the `pytng documentation`_ for further reading
about how pytng works under the hood. Pytng is an optional dependency and must
be installed to use this Reader. Use of the reader without pytng installed will
raise an `ImportError`. The variable `HAS_PYTNG` indicates whether this module
has pytng availble.

In addition to particle-dependent trajectory information like positions,
forces and velocities, the TNG format can store trajectory metadata and
other arbitrary time dependent data. Additional information can range from
the virial and pressure components to the molecular topology of the system.
This is enabled by a block based system in which binary flags indicate the
presence or absence of various data blocks. The structure of a TNG file is
provided  in the `TNG specification`_. The TNG paper :footcite:p:`Lundborg2014` and
the `pytng documentation`_ are also good resources. The user is encouraged to
read the full list of `TNG blocks`_ to understand the full power of the TNG
format.



Current Limitations
-------------------
Currently there is only a Reader for TNG files and no Writer. This will depend
on upstream changes in pytng. Additionally, it is not currently possible to
read the molecular topology from a TNG file.

There are also some limitations to reading TNG files with pytng.
Currently we do not allow data to be read *off stride*. In essence this
means that all of the critical trajectory data (positions, box, velocities
(if present), forces (if present)) must share the same stride in trajectory
integrator steps. These critical blocks in the TNG file are henceforth called
*special blocks*. Optional blocks (all blocks that are not special blocks)
will not be read if they do not share an integrator step with, or are not
divisible by the shared integrator step of the special blocks.


References
----------

.. footbibliography::


.. Links

.. _GROMACS:
    https://www.gromacs.org/
.. _pytng:
    https://github.com/MDAnalysis/pytng
.. _pytng documentation:
    https://www.mdanalysis.org/pytng/
.. _TNG blocks:
    https://www.mdanalysis.org/pytng/documentation_pages/Blocks.html
.. _TNG specification:
    https://gitlab.com/hugomacdermott/tng/-/blob/master/Trajectoryformatspecification.mk

"""

import warnings
from typing import List, Optional

import numpy as np
from MDAnalysis.coordinates import base
from MDAnalysis.coordinates.timestep import Timestep
from MDAnalysis.lib.mdamath import triclinic_box
from MDAnalysis.lib.util import store_init_arguments

from ..due import Doi, due

try:
    import pytng
except ImportError:
    # Indicates whether pytng is found.
    HAS_PYTNG = False
else:
    # Indicates whether pytng is found.
    HAS_PYTNG = True


class TNGReader(base.ReaderBase):
    r"""Reader for the TNG format

    The TNG format :footcite:p:`Lundborg2014` is a format used in `GROMACS`_ for
    storage of trajectory and topology information. This reader is are based on
    the `pytng`_ package for reading TNG files. The reader is directed to the
    `pytng documentation`_ and `TNG specification`_ for further reading.

    The TNG format allows a wide range of compression
    algorithms and unlike the compressed XTC format can also store velocities
    and forces in addition to positions.

    The contents of the *special blocks* (positions, box, velocities, forces)
    are read into the timestep if present. Additional blocks are read into the
    `ts.data` dictionary if they are available at the current frame.

    .. versionadded:: 2.4.0
    """

    format = "TNG"
    # NOTE: Time units are in seconds unlike other GROMACS formats
    units = {
        "time": "second",
        "length": "nm",
        "velocity": "nm/ps",
        "force": "kJ/(mol*nm)",
    }

    _box_blockname = "TNG_TRAJ_BOX_SHAPE"
    _positions_blockname = "TNG_TRAJ_POSITIONS"
    _velocities_blockname = "TNG_TRAJ_VELOCITIES"
    _forces_blockname = "TNG_TRAJ_FORCES"
    _special_blocks = [
        _box_blockname,
        _positions_blockname,
        _velocities_blockname,
        _forces_blockname,
    ]

    @due.dcite(
        Doi("10.1002/jcc.23495"), description="The TNG paper", path=__name__
    )
    @store_init_arguments
    def __init__(self, filename: str, convert_units: bool = True, **kwargs):
        """Initialize a TNG trajectory

        Parameters
        ----------
        filename : str
            filename of the trajectory
        convert_units : bool (optional)
            convert into MDAnalysis units

        """
        if not HAS_PYTNG:
            raise ImportError(
                "TNGReader: To read TNG files please install pytng"
            )

        super(TNGReader, self).__init__(filename, **kwargs)

        self.filename = filename
        self.convert_units = convert_units

        self._file_iterator = pytng.TNGFileIterator(self.filename, "r")
        self.n_atoms = self._file_iterator.n_atoms
        self._n_steps = self._file_iterator.n_steps

        # names of the blocks
        self._block_names = list(self._file_iterator.block_ids.keys())
        # block ids, dict of C long long
        self._block_dictionary = self._file_iterator.block_ids
        self._block_strides = self._file_iterator.block_strides
        self._data_frames = self._file_iterator.n_data_frames
        # init all special blocks to not present
        self._special_block_present = {k: False for k in self._special_blocks}

        self.ts = self._Timestep(self.n_atoms, **self._ts_kwargs)

        # check for all the special blocks
        self._has_box = self._box_blockname in self._block_names
        if self._has_box:
            self._special_block_present[self._box_blockname] = True
            self.ts.dimensions = np.zeros(6, dtype=np.float32)

        self._has_positions = self._positions_blockname in self._block_names
        if self._has_positions:
            self._special_block_present[self._positions_blockname] = True
            self.ts.positions = (
                self._file_iterator.make_ndarray_for_block_from_name(
                    self._positions_blockname
                )
            )

        self._has_velocities = self._velocities_blockname in self._block_names
        if self._has_velocities:
            self._special_block_present[self._velocities_blockname] = True
            self.ts.velocities = (
                self._file_iterator.make_ndarray_for_block_from_name(
                    self._velocities_blockname
                )
            )

        self._has_forces = self._forces_blockname in self._block_names
        if self._has_forces:
            self._special_block_present[self._forces_blockname] = True
            self.ts.forces = (
                self._file_iterator.make_ndarray_for_block_from_name(
                    self._forces_blockname
                )
            )

        # check for any additional blocks that will be read into ts.data
        self._additional_blocks = [
            block
            for block in self._block_names
            if block not in self._special_blocks
        ]
        self._check_strides_and_frames()
        self._frame = 0
        # box needs a temporary place to be stored as ts.dimensions is
        # wrong shape initially
        self._box_temp = self._file_iterator.make_ndarray_for_block_from_name(
            self._box_blockname
        )
        self._frame = -1
        self._read_next_timestep()

    def _check_strides_and_frames(self):
        """
        Check that the strides and frame numbers of the blocks in the TNG
        file match up so that the file can be iterated over retrieving data at
        each integrator step
        """

        strides = []
        n_data_frames = []

        for block in self.special_blocks:
            stride = self._block_strides[block]
            strides.append(stride)
            n_data_frame = self._data_frames[block]
            n_data_frames.append(n_data_frame)

        strides_eq = all(v == strides[0] for v in strides)
        frames_eq = all(v == n_data_frames[0] for v in n_data_frames)

        if (not strides_eq) or (not frames_eq):
            raise IOError(
                "Strides of TNG special blocks not equal, file cannot be read"
            )

        self._global_stride = strides[0]
        # NOTE frame number is 0 indexed so increment
        self._n_frames = n_data_frames[0] + 1

        self._additional_blocks_to_read = []
        for block in self._additional_blocks:
            stride_add = self._block_strides[block]
            if stride_add != self._global_stride:
                if stride_add % self._global_stride:  # pragma: no cover
                    warnings.warn(
                        f"TNG additional block {block} does not match"
                        " strides of other blocks and is not"
                        " divisible by the global stride_length."
                        " It will not be read"
                    )
                else:
                    self._additional_blocks_to_read.append(
                        block
                    )  # pragma: no cover
            else:
                self._additional_blocks_to_read.append(block)

    def close(self):
        """close the reader"""
        self._file_iterator._close()

    @staticmethod
    def parse_n_atoms(filename: str) -> int:
        """parse the number of atoms in the TNG trajectory file

        Parameters
        ----------
        filename : str
            The name of the TNG file

        Returns
        -------
        n_atoms : int
            The number of atoms in the TNG file

        """
        if not HAS_PYTNG:
            raise ImportError(
                "TNGReader: To read TNG files please install pytng"
            )
        with pytng.TNGFileIterator(filename, "r") as tng:
            n_atoms = tng.n_atoms
        return n_atoms

    @property
    def n_frames(self) -> int:
        """number of frames in trajectory

        Returns
        -------
        n_frames : int
            The number of data containing frames in the trajectory
        """
        return self._n_frames

    @property
    def blocks(self) -> List[str]:
        """list of the blocks that are in the file

        Returns
        -------

        blocks : list
            The block names present in the TNG trajectory
        """
        return self._block_names

    @property
    def special_blocks(self) -> List[str]:
        """list of the special blocks that are in the file

        Returns
        -------

        special_blocks : list
            The special block names (positions, box, velocities, forces)
            present in the TNG trajectory
        """
        return [k for k, v in self._special_block_present.items() if v]

    @property
    def additional_blocks(self) -> List[str]:
        """list of the additional (non-special) blocks that are being read from
        the trajectory. This may be exclude some blocks present in the file if
        they do not fall on the same trajectory stride as the positions and
        velocities.

        Returns
        -------

        additional_blocks : list
            The additional block names in the TNG trajectory
        """
        return self._additional_blocks_to_read

    def _reopen(self):
        """reopen the trajectory"""
        self.ts.frame = 0
        self._frame = -1
        self._file_iterator._close()
        self._file_iterator._open(self.filename, "r")

    def _frame_to_step(self, frame: int) -> int:
        """Convert a frame number to an integrator step

        Parameters
        ----------
        frame : int
            The frame number

        Returns
        -------
        integrators_step :  int
            The integrator step of the frame in the TNG file

        """
        return frame * self._global_stride

    def _read_frame(self, i: int) -> Timestep:
        """read frame i

        Parameters
        ----------
        i : int
            The trajectory frame to be read

        Returns
        -------
        ts : Timestep
            Data from frame i encapsulated in an MDA `:class:Timestep`
        """
        self._frame = i - 1
        ts = self._read_next_timestep()
        return ts

    def _read_next_timestep(self, ts: Optional[Timestep] = None) -> Timestep:
        """Read next frame into a timestep

        Parameters
        ----------
        ts : Timestep
            The timestep to read the data into

        Returns
        -------
        ts :  Timestep
            The timestep filled with data from the next step
        """
        if self._frame == self.n_frames - 1:
            raise IOError("trying to go over trajectory limit")
        if ts is None:
            ts = self.ts
        # convert from frames to integrator steps
        step = self._frame_to_step(self._frame + 1)
        iterator_step = self._file_iterator.read_step(step)
        self._frame += 1
        ts = self._frame_to_ts(iterator_step, ts)
        return ts

    def _frame_to_ts(
        self, curr_step: "pytng.TNGCurrentIntegratorStep", ts: Timestep
    ) -> Timestep:
        """convert a TNGCurrentIteratorStep to an MDA Timestep

        Parameters
        ----------
        curr_step : pytng.TNGCurrentIntegratorStep
            The current timestep in the TNG trajectory

        ts : Timestep
            The current timestep in the MDA trajectory

        Returns
        -------
        ts: Timestep
            The current timestep in the MDA trajectory with data from TNG
            trajectory integrator step

        Raises
        ------
        IOError
            The reading of one of the attributes from the TNG file failed
        """

        ts.frame = self._frame
        time = curr_step.get_time()
        if self.convert_units:
            time = self.convert_time_from_native(time)
        ts.time = time
        ts.data["step"] = curr_step.step

        if self._has_box:
            curr_step.get_box(self._box_temp)
            ts.dimensions = triclinic_box(*self._box_temp.reshape(3, 3))
            if not curr_step.read_success:
                raise IOError("Failed to read box from TNG file")
            if self.convert_units:
                self.convert_pos_from_native(ts.dimensions[:3])
        if self._has_positions:
            curr_step.get_positions(ts.positions)
            if not curr_step.read_success:
                raise IOError("Failed to read positions from TNG file")
            if self.convert_units:
                self.convert_pos_from_native(ts.positions)
        if self._has_velocities:
            curr_step.get_velocities(ts.velocities)
            if not curr_step.read_success:
                raise IOError("Failed to read velocities from TNG file")
            if self.convert_units:
                self.convert_velocities_from_native(ts.velocities)
        if self._has_forces:
            curr_step.get_forces(ts.forces)
            if not curr_step.read_success:
                raise IOError("Failed to read forces from TNG file")
            if self.convert_units:
                self.convert_forces_from_native(self.ts.forces)

        for block in self._additional_blocks_to_read:
            add_block_stride = self._block_strides[block]
            # check we are on stride for our block
            if not (add_block_stride % self._global_stride):
                block_data = (
                    self._file_iterator.make_ndarray_for_block_from_name(block)
                )
                # additional blocks read into ts.data dictionary
                ts.data[block] = curr_step.get_blockid(
                    self._block_dictionary[block], block_data
                )
                if not curr_step.read_success:
                    raise IOError(
                        f"Failed to read additional block {block} from TNG file"
                    )
        return ts

    def __getstate__(self):
        """Make a dictionary of the class state to pickle Reader instance.

        Must be done manually as pytng uses a non-trivial`__cinit__`.
        """
        state = self.__dict__.copy()
        # cant have PyTNG file iterator in as it is non-pickable
        del state["_file_iterator"]
        return state

    def __setstate__(self, state):
        """Restore class from `state` dictionary in unpickling of Reader
        instance
        """
        self.__dict__ = state
        # reconstruct file iterator
        self._file_iterator = pytng.TNGFileIterator(self.filename, "r")

        # unlike self._read_frame(self._frame),
        # the following lines update the state of the C-level file iterator
        # without updating the ts object.
        # This is necessary to preserve the modification,
        # e.g. changing coordinates, in the ts object.
        # see PR #3722 for more details.
        step = self._frame_to_step(self._frame)
        _ = self._file_iterator.read_step(step)

    def Writer(self):
        """Writer for TNG files

        Raises
        ------
        NotImplementedError
            Currently there is not writer for TNG files pending implementation
            upstream in pytng.
        """
        raise NotImplementedError("There is currently no writer for TNG files")



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import importlib
import os
import platform
import sys
import tempfile
import warnings
from importlib import reload

import MDAnalysis as mda
import MDAnalysis.analysis.align as align
import MDAnalysis.analysis.encore as encore
import MDAnalysis.analysis.encore.confdistmatrix as confdistmatrix
import MDAnalysis.analysis.rms as rms
import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal

from MDAnalysisTests import block_import
from MDAnalysisTests.datafiles import DCD, DCD2, PSF, TPR, XTC


def function(x):
    return x**2


def test_moved_to_mdakit_warning():
    wmsg = "MDAnalysis.analysis.encore is deprecated"
    with pytest.warns(DeprecationWarning, match=wmsg):
        reload(encore)


class TestEncore(object):
    @pytest.fixture(scope="class")
    def ens1_template(self):
        template = mda.Universe(PSF, DCD)
        template.transfer_to_memory(step=5)
        return template

    @pytest.fixture(scope="class")
    def ens2_template(self):
        template = mda.Universe(PSF, DCD2)
        template.transfer_to_memory(step=5)
        return template

    @pytest.fixture()
    def ens1(self, ens1_template):
        return mda.Universe(
            ens1_template.filename,
            ens1_template.trajectory.timeseries(order="fac"),
            format=mda.coordinates.memory.MemoryReader,
        )

    @pytest.fixture()
    def ens2(self, ens2_template):
        return mda.Universe(
            ens2_template.filename,
            ens2_template.trajectory.timeseries(order="fac"),
            format=mda.coordinates.memory.MemoryReader,
        )

    def test_triangular_matrix(self):
        scalar = 2
        size = 3
        expected_value = 1.984
        filename = tempfile.mktemp() + ".npz"

        triangular_matrix = encore.utils.TriangularMatrix(size=size)

        triangular_matrix[0, 1] = expected_value

        err_msg = (
            "Data error in TriangularMatrix: read/write are not consistent"
        )
        assert_equal(triangular_matrix[0, 1], expected_value, err_msg)

        assert_equal(
            triangular_matrix[0, 1],
            triangular_matrix[1, 0],
            err_msg="Data error in TriangularMatrix: matrix non symmetrical",
        )
        triangular_matrix.savez(filename)

        triangular_matrix_2 = encore.utils.TriangularMatrix(
            size=size, loadfile=filename
        )
        assert_equal(
            triangular_matrix_2[0, 1],
            expected_value,
            err_msg="Data error in TriangularMatrix: loaded matrix non symmetrical",
        )

        triangular_matrix_3 = encore.utils.TriangularMatrix(size=size)
        triangular_matrix_3.loadz(filename)
        assert_equal(
            triangular_matrix_3[0, 1],
            expected_value,
            err_msg="Data error in TriangularMatrix: loaded matrix non symmetrical",
        )

        incremented_triangular_matrix = triangular_matrix + scalar
        assert_equal(
            incremented_triangular_matrix[0, 1],
            expected_value + scalar,
            err_msg="Error in TriangularMatrix: addition of scalar gave"
            "inconsistent results",
        )

        triangular_matrix += scalar
        assert_equal(
            triangular_matrix[0, 1],
            expected_value + scalar,
            err_msg="Error in TriangularMatrix: addition of scalar gave"
            "inconsistent results",
        )

        multiplied_triangular_matrix_2 = triangular_matrix_2 * scalar
        assert_equal(
            multiplied_triangular_matrix_2[0, 1],
            expected_value * scalar,
            err_msg="Error in TriangularMatrix: multiplication by scalar gave"
            "inconsistent results",
        )

        triangular_matrix_2 *= scalar
        assert_equal(
            triangular_matrix_2[0, 1],
            expected_value * scalar,
            err_msg="Error in TriangularMatrix: multiplication by scalar gave\
inconsistent results",
        )

    @pytest.mark.xfail(os.name == "nt", reason="Not yet supported on Windows.")
    def test_parallel_calculation(self):

        arguments = [tuple([i]) for i in np.arange(0, 100)]

        parallel_calculation = encore.utils.ParallelCalculation(
            function=function, n_jobs=4, args=arguments
        )
        results = parallel_calculation.run()

        for i, r in enumerate(results):
            assert_equal(
                r[1],
                arguments[i][0] ** 2,
                err_msg="Unexpected results from ParallelCalculation",
            )

    def test_rmsd_matrix_with_superimposition(self, ens1):
        conf_dist_matrix = (
            encore.confdistmatrix.conformational_distance_matrix(
                ens1,
                encore.confdistmatrix.set_rmsd_matrix_elements,
                select="name CA",
                pairwise_align=True,
                weights="mass",
                n_jobs=1,
            )
        )

        reference = rms.RMSD(ens1, select="name CA")
        reference.run()
        err_msg = (
            "Calculated RMSD values differ from "
            "the reference implementation"
        )
        for i, rmsd in enumerate(reference.results.rmsd):
            assert_allclose(
                conf_dist_matrix[0, i],
                rmsd[2],
                rtol=0,
                atol=1.5e-3,
                err_msg=err_msg,
            )

    def test_rmsd_matrix_with_superimposition_custom_weights(self, ens1):
        conf_dist_matrix = (
            encore.confdistmatrix.conformational_distance_matrix(
                ens1,
                encore.confdistmatrix.set_rmsd_matrix_elements,
                select="name CA",
                pairwise_align=True,
                weights="mass",
                n_jobs=1,
            )
        )

        conf_dist_matrix_custom = (
            encore.confdistmatrix.conformational_distance_matrix(
                ens1,
                encore.confdistmatrix.set_rmsd_matrix_elements,
                select="name CA",
                pairwise_align=True,
                weights=(
                    ens1.select_atoms("name CA").masses,
                    ens1.select_atoms("name CA").masses,
                ),
                n_jobs=1,
            )
        )

        for i in range(conf_dist_matrix_custom.size):
            assert_allclose(
                conf_dist_matrix_custom[0, i],
                conf_dist_matrix[0, i],
                rtol=0,
                atol=1.5e-7,
            )

    def test_rmsd_matrix_without_superimposition(self, ens1):
        selection_string = "name CA"
        selection = ens1.select_atoms(selection_string)
        reference_rmsd = []
        coordinates = ens1.trajectory.timeseries(selection, order="fac")
        for coord in coordinates:
            reference_rmsd.append(
                rms.rmsd(coordinates[0], coord, superposition=False)
            )

        confdist_matrix = encore.confdistmatrix.conformational_distance_matrix(
            ens1,
            encore.confdistmatrix.set_rmsd_matrix_elements,
            select=selection_string,
            pairwise_align=False,
            weights="mass",
            n_jobs=1,
        )

        print(repr(confdist_matrix.as_array()[0, :]))
        assert_allclose(
            confdist_matrix.as_array()[0, :],
            reference_rmsd,
            rtol=0,
            atol=1.5e-3,
            err_msg="calculated RMSD values differ from reference",
        )

    def test_ensemble_superimposition(self):
        aligned_ensemble1 = mda.Universe(PSF, DCD)
        align.AlignTraj(
            aligned_ensemble1,
            aligned_ensemble1,
            select="name CA",
            in_memory=True,
        ).run()
        aligned_ensemble2 = mda.Universe(PSF, DCD)
        align.AlignTraj(
            aligned_ensemble2,
            aligned_ensemble2,
            select="name *",
            in_memory=True,
        ).run()

        rmsfs1 = rms.RMSF(aligned_ensemble1.select_atoms("name *"))
        rmsfs1.run()

        rmsfs2 = rms.RMSF(aligned_ensemble2.select_atoms("name *"))
        rmsfs2.run()

        assert sum(rmsfs1.results.rmsf) > sum(rmsfs2.results.rmsf), (
            "Ensemble aligned on all "
            "atoms should have lower full-atom RMSF than ensemble aligned on only CAs."
        )

    def test_ensemble_superimposition_to_reference_non_weighted(self):
        aligned_ensemble1 = mda.Universe(PSF, DCD)
        align.AlignTraj(
            aligned_ensemble1,
            aligned_ensemble1,
            select="name CA",
            in_memory=True,
        ).run()
        aligned_ensemble2 = mda.Universe(PSF, DCD)
        align.AlignTraj(
            aligned_ensemble2,
            aligned_ensemble2,
            select="name *",
            in_memory=True,
        ).run()

        rmsfs1 = rms.RMSF(aligned_ensemble1.select_atoms("name *"))
        rmsfs1.run()

        rmsfs2 = rms.RMSF(aligned_ensemble2.select_atoms("name *"))
        rmsfs2.run()

        assert sum(rmsfs1.results.rmsf) > sum(rmsfs2.results.rmsf), (
            "Ensemble aligned on all "
            "atoms should have lower full-atom RMSF than ensemble aligned on only CAs."
        )

    def test_covariance_matrix(self, ens1):
        # fmt: off
        reference_cov = np.array([
            [12.9122,-5.2692,3.9016,10.0663,-5.3309,3.8923,8.5037,-5.2017,2.6941],
            [-5.2692,4.1087,-2.4101,-4.5485,3.3954,-2.3245,-3.7343,2.8415,-1.6223],
            [3.9016,-2.4101,3.1800,3.4453,-2.6860,2.2438,2.7751,-2.2523,1.6084],
            [10.0663,-4.5485,3.4453,8.8608,-4.6727,3.3641,7.0106,-4.4986,2.2604],
            [-5.3309,3.3954,-2.6860,-4.6727,4.4627,-2.4233,-3.8304,3.0367,-1.6942],
            [3.8923,-2.3245,2.2438,3.3641,-2.4233,2.6193,2.6908,-2.0252,1.5775],
            [8.5037,-3.7343,2.7751,7.0106,-3.8304,2.6908,6.2861,-3.7138,1.8701],
            [-5.2017,2.8415,-2.2523,-4.4986,3.0367,-2.0252,-3.7138,3.3999,-1.4166],
            [2.6941,-1.6223,1.6084,2.2604,-1.6942,1.5775,1.8701,-1.4166,1.4664]
        ])
        # fmt: on

        covariance = encore.covariance.covariance_matrix(
            ens1,
            select="name CA and resnum 1:3",
            estimator=encore.covariance.shrinkage_covariance_estimator,
        )
        assert_allclose(
            covariance,
            reference_cov,
            rtol=0,
            atol=1.5e-4,
            err_msg="Covariance matrix from covariance estimation not as expected",
        )

    def test_covariance_matrix_with_reference(self, ens1):
        # fmt: off
        reference_cov = np.array([
            [39.0760,-28.5383,29.7761,37.9330,-35.5251,18.9421,30.4334,-31.4829,12.8712],
            [-28.5383,24.1827,-25.5676,-29.0183,30.3511,-15.9598,-22.9298,26.1086,-10.8693],
            [29.7761,-25.5676,28.9796,30.7607,-32.8739,17.7072,24.1689,-28.3557,12.1190],
            [37.9330,-29.0183,30.7607,37.6532,-36.4537,19.2865,29.9841,-32.1404,12.9998],
            [-35.5251,30.3511,-32.8739,-36.4537,38.5711,-20.1190,-28.7652,33.2857,-13.6963],
            [18.9421,-15.9598,17.7072,19.2865,-20.1190,11.4059,15.1244,-17.2695,7.8205],
            [30.4334,-22.9298,24.1689,29.9841,-28.7652,15.1244,24.0514,-25.4106,10.2863],
            [-31.4829,26.1086,-28.3557,-32.1404,33.2857,-17.2695,-25.4106,29.1773,-11.7530],
            [12.8712,-10.8693,12.1190,12.9998,-13.6963,7.8205,10.2863,-11.7530,5.5058]
        ])
        # fmt: on

        covariance = encore.covariance.covariance_matrix(
            ens1,
            select="name CA and resnum 1:3",
            estimator=encore.covariance.shrinkage_covariance_estimator,
            reference=ens1,
        )
        err_msg = (
            "Covariance matrix from covariance estimation not as expected"
        )
        assert_allclose(
            covariance, reference_cov, rtol=0, atol=1.5e-4, err_msg=err_msg
        )

    def test_hes_to_self(self, ens1):
        results, details = encore.hes([ens1, ens1])
        result_value = results[0, 1]
        expected_value = 0.0
        assert_allclose(
            result_value,
            expected_value,
            rtol=0,
            atol=1.5e-7,
            err_msg="Harmonic Ensemble Similarity to itself\
                                 not zero:{0:f}".format(
                result_value
            ),
        )

    def test_hes(self, ens1, ens2):
        results, details = encore.hes([ens1, ens2], weights="mass")
        result_value = results[0, 1]
        min_bound = 1e5
        assert result_value > min_bound, (
            "Unexpected value for Harmonic "
            "Ensemble Similarity: {0:f}. Expected {1:f}.".format(
                result_value, min_bound
            )
        )

    def test_hes_custom_weights(self, ens1, ens2):
        results, details = encore.hes([ens1, ens2], weights="mass")
        results_custom, details_custom = encore.hes(
            [ens1, ens2],
            weights=(
                ens1.select_atoms("name CA").masses,
                ens2.select_atoms("name CA").masses,
            ),
        )
        result_value = results[0, 1]
        result_value_custom = results_custom[0, 1]
        assert_allclose(result_value, result_value_custom, rtol=0, atol=1.5e-7)

    def test_hes_align(self, ens1, ens2):
        # This test is massively sensitive!
        # Get 5260 when masses were float32?
        results, details = encore.hes([ens1, ens2], align=True)
        result_value = results[0, 1]
        expected_value = 2047.05
        assert_allclose(
            result_value,
            expected_value,
            rtol=0,
            atol=1.5e3,
            err_msg="Unexpected value for Harmonic Ensemble Similarity: {0:f}. Expected {1:f}.".format(
                result_value, expected_value
            ),
        )

    def test_ces_to_self(self, ens1):
        results, details = encore.ces(
            [ens1, ens1],
            clustering_method=encore.AffinityPropagationNative(
                preference=-3.0
            ),
        )
        result_value = results[0, 1]
        expected_value = 0.0
        assert_allclose(
            result_value,
            expected_value,
            rtol=0,
            atol=1.5e-7,
            err_msg="ClusteringEnsemble Similarity to itself not zero: {0:f}".format(
                result_value
            ),
        )

    def test_ces(self, ens1, ens2):
        results, details = encore.ces([ens1, ens2])
        result_value = results[0, 1]
        expected_value = 0.51
        assert_allclose(
            result_value,
            expected_value,
            rtol=0,
            atol=1.5e-2,
            err_msg="Unexpected value for Cluster Ensemble Similarity: {0:f}. Expected {1:f}.".format(
                result_value, expected_value
            ),
        )

    def test_dres_to_self(self, ens1):
        results, details = encore.dres([ens1, ens1])
        result_value = results[0, 1]
        expected_value = 0.0
        assert_allclose(
            result_value,
            expected_value,
            rtol=0,
            atol=1.5e-2,
            err_msg="Dim. Reduction Ensemble Similarity to itself not zero: {0:f}".format(
                result_value
            ),
        )

    def test_dres(self, ens1, ens2):
        results, details = encore.dres(
            [ens1, ens2], select="name CA and resnum 1-10"
        )
        result_value = results[0, 1]
        upper_bound = 0.6
        assert result_value < upper_bound, (
            "Unexpected value for Dim. "
            "reduction Ensemble Similarity: {0:f}. Expected {1:f}.".format(
                result_value, upper_bound
            )
        )

    @pytest.mark.xfail  # sporadically fails, see Issue #2158
    def test_dres_without_superimposition(self, ens1, ens2):
        distance_matrix = encore.get_distance_matrix(
            encore.merge_universes([ens1, ens2]), superimpose=False
        )
        results, details = encore.dres(
            [ens1, ens2], distance_matrix=distance_matrix
        )
        result_value = results[0, 1]
        expected_value = 0.68
        assert_allclose(
            result_value,
            expected_value,
            rtol=0,
            atol=1.5e-1,
            err_msg="Unexpected value for Dim. reduction Ensemble Similarity: {0:f}. Expected {1:f}.".format(
                result_value, expected_value
            ),
        )

    def test_ces_convergence(self, ens1):
        expected_values = [0.3443593, 0.1941854, 0.06857104, 0.0]
        results = encore.ces_convergence(ens1, 5)
        for i, ev in enumerate(expected_values):
            assert_allclose(
                ev,
                results[i],
                rtol=0,
                atol=1.5e-2,
                err_msg="Unexpected value for Clustering Ensemble similarity in convergence estimation",
            )

    def test_dres_convergence(self, ens1):
        # Due to encore.dres_convergence() involving random numbers, the
        # following assertion is allowed to fail once. This significantly
        # reduces the probability of a random test failure.
        expected_values = [0.3, 0.0]
        results = encore.dres_convergence(ens1, 10)
        try:
            assert_allclose(
                results[:, 0], expected_values, rtol=0, atol=1.5e-1
            )
        except AssertionError:
            # Random test failure is very rare, but repeating the failed test
            # just once would only assert that the test passes with 50%
            # probability. To be a little safer, we raise a warning and repeat
            # the test 10 times:
            warnings.warn(
                message="Test 'test_dres_convergence' failed, "
                "repeating test 10 times.",
                category=RuntimeWarning,
            )
            for i in range(10):
                results = encore.dres_convergence(ens1, 10)
                assert_allclose(
                    results[:, 0],
                    expected_values,
                    rtol=0,
                    atol=1.5e-1,
                    err_msg="Unexpected value for Dim. "
                    "reduction Ensemble similarity in "
                    "convergence estimation",
                )

    @pytest.mark.xfail  # sporadically fails, see Issue #2158
    def test_hes_error_estimation(self, ens1):
        expected_average = 10
        expected_stdev = 12
        averages, stdevs = encore.hes(
            [ens1, ens1],
            estimate_error=True,
            bootstrapping_samples=10,
            select="name CA and resnum 1-10",
        )
        average = averages[0, 1]
        stdev = stdevs[0, 1]
        err_msg = (
            "Unexpected average value for bootstrapped samples in Harmonic"
            " Ensemble similarity"
        )
        error_msg = (
            "Unexpected standard deviation for bootstrapped samples in"
            " Harmonic Ensemble similarity"
        )
        assert_allclose(
            average, expected_average, rtol=0, atol=1.5e2, err_msg=err_msg
        )
        assert_allclose(
            stdev, expected_stdev, rtol=0, atol=1.5e2, err_msg=error_msg
        )

    def test_ces_error_estimation(self, ens1):
        expected_average = 0.03
        expected_stdev = 0.31
        averages, stdevs = encore.ces(
            [ens1, ens1],
            estimate_error=True,
            bootstrapping_samples=10,
            clustering_method=encore.AffinityPropagationNative(
                preference=-2.0
            ),
            select="name CA and resnum 1-10",
        )
        average = averages[0, 1]
        stdev = stdevs[0, 1]

        assert_allclose(
            average,
            expected_average,
            rtol=0,
            atol=1.5e-1,
            err_msg="Unexpected average value for bootstrapped samples in Clustering Ensemble similarity",
        )
        assert_allclose(
            stdev,
            expected_stdev,
            rtol=0,
            atol=1.5,
            err_msg="Unexpected standard deviation  for bootstrapped samples in Clustering Ensemble similarity",
        )

    def test_ces_error_estimation_ensemble_bootstrap(self, ens1):
        # Error estimation using a method that does not take a distance
        # matrix as input, and therefore relies on bootstrapping the ensembles
        # instead

        pytest.importorskip("sklearn")

        expected_average = 0.03
        expected_stdev = 0.02
        averages, stdevs = encore.ces(
            [ens1, ens1],
            estimate_error=True,
            bootstrapping_samples=10,
            clustering_method=encore.KMeans(n_clusters=2),
            select="name CA and resnum 1-10",
        )
        average = averages[0, 1]
        stdev = stdevs[0, 1]
        err_msg = (
            "Unexpected average value for bootstrapped samples in"
            " Clustering Ensemble similarity"
        )
        assert_allclose(
            average, expected_average, rtol=0, atol=1.5e-1, err_msg=err_msg
        )
        error_msg = (
            "Unexpected standard deviation for bootstrapped samples in"
            " Clustering Ensemble similarity"
        )
        assert_allclose(
            stdev, expected_stdev, rtol=0, atol=1.5e-1, err_msg=error_msg
        )

    def test_dres_error_estimation(self, ens1):
        average_upper_bound = 0.3
        stdev_upper_bound = 0.2
        averages, stdevs = encore.dres(
            [ens1, ens1],
            estimate_error=True,
            bootstrapping_samples=10,
            select="name CA and resnum 1-10",
        )
        average = averages[0, 1]
        stdev = stdevs[0, 1]
        err_msg = (
            "Unexpected average value for bootstrapped samples in Dim. "
            "reduction Ensemble similarity"
        )
        error_msg = (
            "Unexpected standard deviation for bootstrapped samples in Dim. "
            "reduction Ensemble similarity"
        )
        assert average < average_upper_bound, err_msg
        assert stdev < stdev_upper_bound, error_msg


class TestEncoreClustering(object):
    @pytest.fixture(scope="class")
    def ens1_template(self):
        template = mda.Universe(PSF, DCD)
        template.transfer_to_memory(step=5)
        return template

    @pytest.fixture(scope="class")
    def ens2_template(self):
        template = mda.Universe(PSF, DCD2)
        template.transfer_to_memory(step=5)
        return template

    @pytest.fixture(scope="class")
    def cc(self):
        return encore.ClusterCollection([1, 1, 1, 3, 3, 5, 5, 5])

    @pytest.fixture(scope="class")
    def cluster(self):
        return encore.Cluster(elem_list=np.array([0, 1, 2]), centroid=1)

    @pytest.fixture()
    def ens1(self, ens1_template):
        return mda.Universe(
            ens1_template.filename,
            ens1_template.trajectory.timeseries(order="fac"),
            format=mda.coordinates.memory.MemoryReader,
        )

    @pytest.fixture()
    def ens2(self, ens2_template):
        return mda.Universe(
            ens2_template.filename,
            ens2_template.trajectory.timeseries(order="fac"),
            format=mda.coordinates.memory.MemoryReader,
        )

    def test_clustering_one_ensemble(self, ens1):
        cluster_collection = encore.cluster(ens1)
        expected_value = 7
        assert (
            len(cluster_collection) == expected_value
        ), "Unexpected " "results: {0}".format(cluster_collection)

    def test_clustering_two_ensembles(self, ens1, ens2):
        cluster_collection = encore.cluster([ens1, ens2])
        expected_value = 14
        assert (
            len(cluster_collection) == expected_value
        ), "Unexpected " "results: {0}".format(cluster_collection)

    @pytest.mark.xfail(
        platform.machine() == "arm64" and platform.system() == "Darwin",
        reason="see gh-3599",
    )
    def test_clustering_three_ensembles_two_identical(self, ens1, ens2):
        cluster_collection = encore.cluster([ens1, ens2, ens1])
        expected_value = 40
        assert (
            len(cluster_collection) == expected_value
        ), "Unexpected result:" " {0}".format(cluster_collection)

    def test_clustering_two_methods(self, ens1):
        cluster_collection = encore.cluster(
            [ens1],
            method=[
                encore.AffinityPropagationNative(),
                encore.AffinityPropagationNative(),
            ],
        )
        assert len(cluster_collection[0]) == len(
            cluster_collection[1]
        ), "Unexpected result: {0}".format(cluster_collection)

    def test_clustering_AffinityPropagationNative_direct(self, ens1):
        method = encore.AffinityPropagationNative()
        distance_matrix = encore.get_distance_matrix(ens1)
        cluster_assignment = method(distance_matrix)
        expected_value = 7
        assert (
            len(set(cluster_assignment)) == expected_value
        ), "Unexpected result: {0}".format(cluster_assignment)

    def test_clustering_AffinityPropagation_direct(self, ens1):
        pytest.importorskip("sklearn")
        method = encore.AffinityPropagation(random_state=0)
        distance_matrix = encore.get_distance_matrix(ens1)
        cluster_assignment = method(distance_matrix)
        expected_value = 7
        assert (
            len(set(cluster_assignment)) == expected_value
        ), "Unexpected result: {0}".format(cluster_assignment)

    def test_clustering_KMeans_direct(self, ens1):
        pytest.importorskip("sklearn")
        clusters = 10
        method = encore.KMeans(clusters)
        coordinates = ens1.trajectory.timeseries(order="fac")
        coordinates = np.reshape(coordinates, (coordinates.shape[0], -1))
        cluster_assignment = method(coordinates)
        assert (
            len(set(cluster_assignment)) == clusters
        ), "Unexpected result: {0}".format(cluster_assignment)

    def test_clustering_DBSCAN_direct(self, ens1):
        pytest.importorskip("sklearn")
        method = encore.DBSCAN(eps=0.5, min_samples=2)
        distance_matrix = encore.get_distance_matrix(ens1)
        cluster_assignment = method(distance_matrix)
        expected_value = 2
        assert (
            len(set(cluster_assignment)) == expected_value
        ), "Unexpected result: {0}".format(cluster_assignment)

    def test_clustering_two_different_methods(self, ens1):
        pytest.importorskip("sklearn")
        cluster_collection = encore.cluster(
            [ens1],
            method=[
                encore.AffinityPropagation(preference=-7.5, random_state=0),
                encore.DBSCAN(min_samples=2),
            ],
        )
        assert len(cluster_collection[0]) == len(
            cluster_collection[1]
        ), "Unexpected result: {0}".format(cluster_collection)

    def test_clustering_method_w_no_distance_matrix(self, ens1):
        pytest.importorskip("sklearn")
        cluster_collection = encore.cluster([ens1], method=encore.KMeans(10))
        assert len(cluster_collection) == 10, "Unexpected result: {0}".format(
            cluster_collection
        )

    def test_clustering_two_methods_one_w_no_distance_matrix(self, ens1):
        pytest.importorskip("sklearn")
        cluster_collection = encore.cluster(
            [ens1],
            method=[encore.KMeans(17), encore.AffinityPropagationNative()],
        )
        assert len(cluster_collection[0]) == len(
            cluster_collection[0]
        ), "Unexpected result: {0}".format(cluster_collection)

    def test_sklearn_affinity_propagation(self, ens1):
        pytest.importorskip("sklearn")
        cc1 = encore.cluster([ens1])
        cc2 = encore.cluster(
            [ens1], method=encore.AffinityPropagation(random_state=0)
        )
        assert len(cc1) == len(cc2), (
            "Native and sklearn implementations of affinity "
            "propagation don't agree: mismatch in number of "
            "clusters: {0} {1}".format(len(cc1), len(cc2))
        )

    def test_ClusterCollection_init(self, cc):
        err_msg = "ClusterCollection was not constructed correctly"
        assert_equal(cc.clusters[0].elements, [0, 1, 2], err_msg)
        assert_equal(cc.clusters[1].elements, [3, 4], err_msg)
        assert_equal(cc.clusters[2].elements, [5, 6, 7], err_msg)
        assert_equal(cc.clusters[1].centroid, 3, err_msg)
        assert_equal(cc.clusters[2].centroid, 5, err_msg)

    def test_Cluster_init(self, cluster):
        err_msg = "Cluster was not constructed correctly"
        assert_equal(cluster.elements, [0, 1, 2], err_msg)
        assert_equal(cluster.centroid, 1, err_msg)

    def test_ClusterCollection_get_ids(self, cc):
        assert_equal(
            cc.get_ids(),
            [0, 1, 2],
            err_msg="ClusterCollection ids aren't as expected",
        )

    def test_ClusterCollection_get_centroids(self, cc):
        assert_equal(
            cc.get_centroids(),
            [1, 3, 5],
            err_msg="ClusterCollection centroids aren't as expected",
        )

    def test_cluster_add_metadata(self, cluster):
        metadata = cluster.elements * 10
        cluster.add_metadata("test", metadata)
        assert_equal(
            cluster.metadata["test"],
            metadata,
            err_msg="Cluster metadata isn't as expected",
        )
        metadata = np.append(metadata, 9)
        error_message = (
            "Size of metadata is not equal to the "
            "number of cluster elements"
        )
        with pytest.raises(TypeError, match=error_message):
            cluster.add_metadata("test2", metadata)

    def test_empty_cluster(self):
        empty_cluster = encore.Cluster()
        assert empty_cluster.size == 0
        assert np.size(empty_cluster.elements) == 0
        assert empty_cluster.centroid is None
        assert bool(empty_cluster.metadata) is False

    def test_centroid_not_in_elements(self):
        error_message = "Centroid of cluster not found in the element list"
        with pytest.raises(LookupError, match=error_message):
            encore.Cluster([38, 39, 40, 41, 42, 43], 99)

    def test_metadata_size_error(self):
        error_message = (
            'Size of metadata having label "label" is '
            "not equal to the number of cluster elements"
        )
        with pytest.raises(TypeError, match=error_message):
            encore.Cluster(
                np.array([1, 1, 1]), 1, None, {"label": [1, 1, 1, 1]}
            )

    def test_cluster_iteration(self, cluster):
        test = []
        for i in cluster.elements:
            test.append(i)
        assert_equal(cluster.elements, test)

    def test_cluster_len(self, cluster):
        assert cluster.size == len(cluster)

    def test_cluster_repr(self):
        repr_message = "<Cluster with no elements>"
        assert_equal(repr(encore.Cluster()), repr_message)
        cluster = encore.Cluster(np.array([1]), 1, 1)
        repr_message = "<Cluster with 1 elements, centroid=1, id=1>"
        assert_equal(repr(cluster), repr_message)


class TestEncoreClusteringSklearn(object):
    """The tests in this class were duplicated from the affinity propagation
    tests in scikit-learn"""

    n_clusters = 3

    @pytest.fixture()
    def distance_matrix(self):
        X = np.array(
            [
                [8.73101582, 8.85617874],
                [11.61311169, 11.58774351],
                [10.86083514, 11.06253959],
                [9.45576027, 8.50606967],
                [11.30441509, 11.04867001],
                [8.63708065, 9.02077816],
                [8.34792066, 9.1851129],
                [11.06197897, 11.15126501],
                [11.24563175, 9.36888267],
                [10.83455241, 8.70101808],
                [11.49211627, 11.48095194],
                [10.6448857, 10.20768141],
                [10.491806, 9.38775868],
                [11.08330999, 9.39065561],
                [10.83872922, 9.48897803],
                [11.37890079, 8.93799596],
                [11.70562094, 11.16006288],
                [10.95871246, 11.1642394],
                [11.59763163, 10.91793669],
                [11.05761743, 11.5817094],
                [8.35444086, 8.91490389],
                [8.79613913, 8.82477028],
                [11.00420001, 9.7143482],
                [11.90790185, 10.41825373],
                [11.39149519, 11.89635728],
                [8.31749192, 9.78031016],
                [11.59530088, 9.75835567],
                [11.17754529, 11.13346973],
                [11.01830341, 10.92512646],
                [11.75326028, 8.46089638],
                [11.74702358, 9.36241786],
                [10.53075064, 9.77744847],
                [8.67474149, 8.30948696],
                [11.05076484, 9.16079575],
                [8.79567794, 8.52774713],
                [11.18626498, 8.38550253],
                [10.57169895, 9.42178069],
                [8.65168114, 8.76846013],
                [11.12522708, 10.6583617],
                [8.87537899, 9.02246614],
                [9.29163622, 9.05159316],
                [11.38003537, 10.93945712],
                [8.74627116, 8.85490353],
                [10.65550973, 9.76402598],
                [8.49888186, 9.31099614],
                [8.64181338, 9.154761],
                [10.84506927, 10.8790789],
                [8.98872711, 9.17133275],
                [11.7470232, 10.60908885],
                [10.89279865, 9.32098256],
                [11.14254656, 9.28262927],
                [9.02660689, 9.12098876],
                [9.16093666, 8.72607596],
                [11.47151183, 8.92803007],
                [11.76917681, 9.59220592],
                [9.97880407, 11.26144744],
                [8.58057881, 8.43199283],
                [10.53394006, 9.36033059],
                [11.34577448, 10.70313399],
                [9.07097046, 8.83928763],
            ]
        )

        XX = np.einsum("ij,ij->i", X, X)[:, np.newaxis]
        YY = XX.T
        distances = np.dot(X, X.T)
        distances *= -2
        distances += XX
        distances += YY
        np.maximum(distances, 0, out=distances)
        distances.flat[:: distances.shape[0] + 1] = 0.0
        dimension = len(distances)

        distance_matrix = encore.utils.TriangularMatrix(len(distances))
        for i in range(dimension):
            for j in range(i, dimension):
                distance_matrix[i, j] = distances[i, j]
        return distance_matrix

    def test_one(self, distance_matrix):
        preference = -float(np.median(distance_matrix.as_array()) * 10.0)
        clustering_method = encore.AffinityPropagationNative(
            preference=preference
        )
        ccs = encore.cluster(
            None, distance_matrix=distance_matrix, method=clustering_method
        )
        assert self.n_clusters == len(ccs), (
            "Basic clustering test failed to give the right"
            "number of clusters: {0} vs {1}".format(self.n_clusters, len(ccs))
        )


class TestEncoreDimensionalityReduction(object):
    @pytest.fixture(scope="class")
    def ens1_template(self):
        template = mda.Universe(PSF, DCD)
        template.transfer_to_memory(step=5)
        return template

    @pytest.fixture(scope="class")
    def ens2_template(self):
        template = mda.Universe(PSF, DCD2)
        template.transfer_to_memory(step=5)
        return template

    @pytest.fixture()
    def ens1(self, ens1_template):
        return mda.Universe(
            ens1_template.filename,
            ens1_template.trajectory.timeseries(order="fac"),
            format=mda.coordinates.memory.MemoryReader,
        )

    @pytest.fixture()
    def ens2(self, ens2_template):
        return mda.Universe(
            ens2_template.filename,
            ens2_template.trajectory.timeseries(order="fac"),
            format=mda.coordinates.memory.MemoryReader,
        )

    def test_dimensionality_reduction_one_ensemble(self, ens1):
        dimension = 2
        coordinates, details = encore.reduce_dimensionality(ens1)
        assert_equal(
            coordinates.shape[0],
            dimension,
            err_msg="Unexpected result in dimensionality reduction: {0}".format(
                coordinates
            ),
        )

    def test_dimensionality_reduction_two_ensembles(self, ens1, ens2):
        dimension = 2
        coordinates, details = encore.reduce_dimensionality([ens1, ens2])
        assert_equal(
            coordinates.shape[0],
            dimension,
            err_msg="Unexpected result in dimensionality reduction: {0}".format(
                coordinates
            ),
        )

    def test_dimensionality_reduction_three_ensembles_two_identical(
        self, ens1, ens2
    ):
        coordinates, details = encore.reduce_dimensionality([ens1, ens2, ens1])
        coordinates_ens1 = coordinates[
            :, np.where(details["ensemble_membership"] == 1)
        ]
        coordinates_ens3 = coordinates[
            :, np.where(details["ensemble_membership"] == 3)
        ]
        assert_allclose(
            coordinates_ens1,
            coordinates_ens3,
            rtol=0,
            atol=1.5,
            err_msg="Unexpected result in dimensionality reduction: {0}".format(
                coordinates
            ),
        )

    def test_dimensionality_reduction_specified_dimension(self, ens1, ens2):
        dimension = 3
        coordinates, details = encore.reduce_dimensionality(
            [ens1, ens2],
            method=encore.StochasticProximityEmbeddingNative(
                dimension=dimension
            ),
        )
        assert_equal(
            coordinates.shape[0],
            dimension,
            err_msg="Unexpected result in dimensionality reduction: {0}".format(
                coordinates
            ),
        )

    def test_dimensionality_reduction_SPENative_direct(self, ens1):
        dimension = 2
        method = encore.StochasticProximityEmbeddingNative(dimension=dimension)
        distance_matrix = encore.get_distance_matrix(ens1)
        coordinates, details = method(distance_matrix)
        assert_equal(
            coordinates.shape[0],
            dimension,
            err_msg="Unexpected result in dimensionality reduction: {0}".format(
                coordinates
            ),
        )

    def test_dimensionality_reduction_PCA_direct(self, ens1):
        pytest.importorskip("sklearn")
        dimension = 2
        method = encore.PrincipalComponentAnalysis(dimension=dimension)
        coordinates = ens1.trajectory.timeseries(order="fac")
        coordinates = np.reshape(coordinates, (coordinates.shape[0], -1))
        coordinates, details = method(coordinates)
        assert_equal(
            coordinates.shape[0],
            dimension,
            err_msg="Unexpected result in dimensionality reduction: {0}".format(
                coordinates
            ),
        )

    def test_dimensionality_reduction_different_method(self, ens1, ens2):
        pytest.importorskip("sklearn")
        dimension = 3
        coordinates, details = encore.reduce_dimensionality(
            [ens1, ens2],
            method=encore.PrincipalComponentAnalysis(dimension=dimension),
        )
        assert_equal(
            coordinates.shape[0],
            dimension,
            err_msg="Unexpected result in dimensionality reduction: {0}".format(
                coordinates
            ),
        )

    def test_dimensionality_reduction_two_methods(self, ens1, ens2):
        dims = [2, 3]
        coordinates, details = encore.reduce_dimensionality(
            [ens1, ens2],
            method=[
                encore.StochasticProximityEmbeddingNative(dims[0]),
                encore.StochasticProximityEmbeddingNative(dims[1]),
            ],
        )
        assert_equal(coordinates[1].shape[0], dims[1])

    def test_dimensionality_reduction_two_different_methods(self, ens1, ens2):
        pytest.importorskip("sklearn")
        dims = [2, 3]
        coordinates, details = encore.reduce_dimensionality(
            [ens1, ens2],
            method=[
                encore.StochasticProximityEmbeddingNative(dims[0]),
                encore.PrincipalComponentAnalysis(dims[1]),
            ],
        )
        assert_equal(coordinates[1].shape[0], dims[1])


class TestEncoreConfDistMatrix(object):
    def test_get_distance_matrix(self):
        # Issue #1324
        u = mda.Universe(TPR, XTC)
        dm = confdistmatrix.get_distance_matrix(u)


class TestEncoreImportWarnings(object):
    @block_import("sklearn")
    def _check_sklearn_import_warns(self, package, recwarn):
        for mod in list(sys.modules):  # list as we're changing as we iterate
            if "encore" in mod:
                sys.modules.pop(mod, None)
        warnings.simplefilter("always")
        # assert_warns(ImportWarning, importlib.import_module, package)
        importlib.import_module(package)
        assert recwarn.pop(ImportWarning)

    def test_import_warnings(self, recwarn):
        for mod in list(sys.modules):  # list as we're changing as we iterate
            if "encore" in mod:
                sys.modules.pop(mod, None)
        for pkg in (
            "MDAnalysis.analysis.encore.dimensionality_reduction.DimensionalityReductionMethod",
            "MDAnalysis.analysis.encore.clustering.ClusteringMethod",
        ):
            self._check_sklearn_import_warns(pkg, recwarn)
            # This is a quickfix! Convert this to a parametrize call in future.



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
XYZ Topology Parser
===================

.. versionadded:: 0.9.1

Reads an xyz file and pulls the atom information from it.  Because
xyz only has atom name information, all information about residues
and segments won't be populated.

Classes
-------

.. autoclass:: XYZParser
   :members:

"""
import numpy as np

from ..lib.util import openany
from .base import TopologyReaderBase
from ..core.topology import Topology
from ..core.topologyattrs import (
    Atomnames,
    Atomids,
    Resids,
    Resnums,
    Segids,
    Elements,
)


class XYZParser(TopologyReaderBase):
    """Parse a list of atoms from an XYZ file.

    Creates the following attributes:
     - Atomnames

    .. note::

        By default, atomtypes and masses will be guessed on Universe creation.
        This may change in release 3.0.
        See :ref:`Guessers` for more information.


    .. versionadded:: 0.9.1

    .. versionchanged: 1.0.0
       Store elements attribute, based on XYZ atom names
     .. versionchanged:: 2.8.0
        Removed type and mass guessing (attributes guessing takes place now
        through universe.guess_TopologyAttrs() API).

    """

    format = "XYZ"

    def parse(self, **kwargs):
        """Read the file and return the structure.

        Returns
        -------
        MDAnalysis Topology object
        """
        with openany(self.filename) as inf:
            natoms = int(inf.readline().strip())
            inf.readline()

            names = np.zeros(natoms, dtype=object)

            # Can't infinitely read as XYZ files can be multiframe
            for i in range(natoms):
                name = inf.readline().split()[0]
                names[i] = name

        attrs = [
            Atomnames(names),
            Atomids(np.arange(natoms) + 1),
            Resids(np.array([1])),
            Resnums(np.array([1])),
            Segids(np.array(["SYSTEM"], dtype=object)),
            Elements(names),
        ]

        top = Topology(natoms, 1, 1, attrs=attrs)

        return top



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np

import MDAnalysis as mda
import pytest
from MDAnalysis.coordinates.memory import MemoryReader
from MDAnalysisTests.datafiles import DCD, PSF
from MDAnalysisTests.coordinates.base import (
    BaseReference,
    MultiframeReaderTest,
)
from MDAnalysis.coordinates.memory import Timestep
from numpy.testing import assert_equal, assert_almost_equal


class MemoryReference(BaseReference):
    def __init__(self):
        super(MemoryReference, self).__init__()

        self.topology = PSF
        self.trajectory = DCD
        self.universe = mda.Universe(PSF, DCD)

        self.n_atoms = self.universe.trajectory.n_atoms
        self.n_frames = self.universe.trajectory.n_frames

        self.dt = self.universe.trajectory.ts.dt
        self.dimensions = self.universe.trajectory.ts.dimensions
        self.totaltime = self.universe.trajectory.totaltime
        self.volume = self.universe.trajectory.ts.volume

        self.first_frame = Timestep(self.n_atoms)
        self.first_frame.positions = np.array(self.universe.trajectory[0])
        self.first_frame.frame = 0
        self.first_frame.time = self.first_frame.frame * self.dt

        self.second_frame = Timestep(self.n_atoms)
        self.second_frame.positions = np.array(self.universe.trajectory[1])
        self.second_frame.frame = 1
        self.second_frame.time = self.second_frame.frame * self.dt

        self.last_frame = Timestep(self.n_atoms)
        self.last_frame.positions = np.array(
            self.universe.trajectory[self.n_frames - 1]
        )
        self.last_frame.frame = self.n_frames - 1
        self.last_frame.time = self.last_frame.frame * self.dt

        self.jump_to_frame = self.first_frame.copy()
        self.jump_to_frame.positions = np.array(self.universe.trajectory[3])
        self.jump_to_frame.frame = 3
        self.jump_to_frame.time = self.jump_to_frame.frame * self.dt

    def reader(self, trajectory):
        return mda.Universe(
            self.topology, trajectory, in_memory=True
        ).trajectory

    def iter_ts(self, i):
        ts = self.universe.trajectory[i]
        # correct time because memory reader doesn't read the correct time
        ts.time = ts.frame * self.dt
        return ts


class TestMemoryReader(MultiframeReaderTest):
    @staticmethod
    @pytest.fixture(scope="class")
    def ref():
        return MemoryReference()

    def test_filename_transefer_to_memory(self):
        # MemoryReader should have a filename attribute set to the trajaectory filename
        universe = mda.Universe(PSF, DCD)
        universe.transfer_to_memory()
        assert_equal(universe.trajectory.filename, DCD)

    def test_filename_array(self):
        # filename attribute of MemoryReader should be None when generated from an array
        universe = mda.Universe(PSF, DCD)
        coordinates = universe.trajectory.timeseries(universe.atoms)
        universe2 = mda.Universe(
            PSF, coordinates, format=MemoryReader, order="afc"
        )
        assert universe2.trajectory.filename is None

    def test_default_memory_layout(self):
        universe1 = mda.Universe(PSF, DCD, in_memory=True)
        universe2 = mda.Universe(PSF, DCD, in_memory=True, order="fac")
        assert_equal(
            universe1.trajectory.get_array().shape,
            universe2.trajectory.get_array().shape,
        )

    def test_iteration(self, ref, reader):
        frames = 0
        for i, frame in enumerate(reader):
            frames += 1
        assert frames == ref.n_frames

    def test_extract_array_afc(self, reader):
        assert_equal(reader.timeseries(order="afc").shape, (3341, 98, 3))

    def test_extract_array_afc(self, reader):
        assert_equal(reader.timeseries(order="afc").shape, (3341, 98, 3))

    def test_extract_array_fac(self, reader):
        assert_equal(reader.timeseries(order="fac").shape, (98, 3341, 3))

    def test_extract_array_cfa(self, reader):
        assert_equal(reader.timeseries(order="cfa").shape, (3, 98, 3341))

    def test_extract_array_acf(self, reader):
        assert_equal(reader.timeseries(order="acf").shape, (3341, 3, 98))

    def test_extract_array_fca(self, reader):
        assert_equal(reader.timeseries(order="fca").shape, (98, 3, 3341))

    def test_extract_array_caf(self, reader):
        assert_equal(reader.timeseries(order="caf").shape, (3, 3341, 98))

    def test_timeseries_skip1(self, ref, reader):
        assert_equal(
            reader.timeseries(ref.universe.atoms).shape, (3341, 98, 3)
        )

    def test_timeseries_skip10(self, reader):
        # Check that timeseries skip works similar to numpy slicing
        array1 = reader.timeseries(step=10)
        array2 = reader.timeseries()[:, ::10, :]
        assert_equal(array1, array2)

    def test_timeseries_view(self, reader):
        # timeseries() is expected to provide a view of the underlying array
        assert reader.timeseries().base is reader.get_array()

    def test_timeseries_subarray_view(self, reader):
        # timeseries() is expected to provide a view of the underlying array
        # also in the case where we slice the array using the start, stop and
        # step options.
        assert (
            reader.timeseries(start=5, stop=15, step=2, order="fac").base
            is reader.get_array()
        )

    def test_timeseries_view_from_universe_atoms(self, ref, reader):
        # timeseries() is expected to provide a view of the underlying array
        # also in the special case when asel=universe.atoms.
        selection = ref.universe.atoms
        assert reader.timeseries(asel=selection).base is reader.get_array()

    def test_timeseries_view_from_select_all(self, ref, reader):
        # timeseries() is expected to provide a view of the underlying array
        # also in the special case when using "all" in selections.
        selection = ref.universe.select_atoms("all")
        assert_equal(
            reader.timeseries(asel=selection).base is reader.get_array(), True
        )

    def test_timeseries_noview(self, ref, reader):
        # timeseries() is expected NOT to provide a view of the underlying array
        # for any other selection than "all".
        selection = ref.universe.select_atoms("name CA")
        assert reader.timeseries(asel=selection).base is not reader.get_array()

    def test_repr(self, reader):
        str_rep = str(reader)
        expected = "<MemoryReader with 98 frames of 3341 atoms>"
        assert_equal(str_rep, expected)

    def test_get_writer_1(self):
        pass

    def test_get_writer_2(self):
        pass

    def test_float32(self, ref):
        # Check that we get float32 positions even when initializing with float64
        coordinates = np.random.uniform(
            size=(100, ref.universe.atoms.n_atoms, 3)
        ).cumsum(0)
        universe = mda.Universe(
            ref.universe.filename, coordinates, format=MemoryReader
        )
        assert_equal(
            universe.trajectory.get_array().dtype, np.dtype("float32")
        )

    def test_position_assignation(self, reader):
        # When coordinates are assigned to a timestep, is the change persistent?
        new_positions = np.ones_like(reader.ts.positions, dtype=np.float32)
        reader.ts.positions = new_positions
        reader[0]
        assert_almost_equal(reader.ts.positions, new_positions)

    def test_timeseries_warns_deprecation(self, reader):
        with pytest.warns(
            DeprecationWarning, match="MemoryReader.timeseries " "inclusive"
        ):
            reader.timeseries(start=0, stop=3, step=1)

    def test_timeseries_asel_warns_deprecation(self, ref, reader):
        selection = ref.universe.atoms
        with pytest.warns(DeprecationWarning, match="asel argument to"):
            reader.timeseries(asel=selection)

    def test_timeseries_atomgroup(self, ref, reader):
        selection = ref.universe.atoms
        reader.timeseries(atomgroup=selection)

    def test_timeseries_atomgroup_asel_mutex(self, ref, reader):
        selection = ref.universe.atoms
        with pytest.raises(ValueError, match="Cannot provide both"):
            reader.timeseries(atomgroup=selection, asel=selection)


class TestMemoryReaderVelsForces(object):
    @staticmethod
    @pytest.fixture(params=["2d", "3d"])
    def ref_pos(request):
        if request.param == "2d":
            return np.arange(30).reshape(10, 3)
        elif request.param == "3d":
            return np.arange(30).reshape(1, 10, 3)

    @staticmethod
    @pytest.fixture(params=["2d", "3d"])
    def ref_vels(request):
        if request.param == "2d":
            return np.arange(30).reshape(10, 3) + 100
        elif request.param == "3d":
            return np.arange(30).reshape(1, 10, 3) + 100

    @staticmethod
    @pytest.fixture(params=["2d", "3d"])
    def ref_forces(request):
        if request.param == "2d":
            return np.arange(30).reshape(10, 3) + 1000
        elif request.param == "3d":
            return np.arange(30).reshape(1, 10, 3) + 1000

    @staticmethod
    def assert_equal_dims(arr1, arr2):
        if arr2.ndim == 3:
            assert_equal(arr1, arr2[0])
        elif arr2.ndim == 2:
            assert_equal(arr1, arr2)

    def test_velocities(self, ref_pos, ref_vels):
        mr = MemoryReader(ref_pos, velocities=ref_vels)

        assert mr.ts.has_velocities
        self.assert_equal_dims(mr.ts.velocities, ref_vels)
        assert not mr.ts.has_forces

    def test_forces(self, ref_pos, ref_forces):
        mr = MemoryReader(ref_pos, forces=ref_forces)

        assert not mr.ts.has_velocities
        assert mr.ts.has_forces
        self.assert_equal_dims(mr.ts.forces, ref_forces)

    def test_both(self, ref_pos, ref_vels, ref_forces):
        mr = MemoryReader(ref_pos, velocities=ref_vels, forces=ref_forces)
        assert mr.ts.has_velocities
        self.assert_equal_dims(mr.ts.velocities, ref_vels)
        assert mr.ts.has_forces
        self.assert_equal_dims(mr.ts.forces, ref_forces)

    @pytest.mark.parametrize("param", ["velocities", "forces"])
    def test_wrongshape(self, ref_pos, param):
        with pytest.raises(ValueError):
            mr = MemoryReader(ref_pos, **{param: np.zeros((3, 2, 1))})


class TestDimensions(object):
    @staticmethod
    @pytest.fixture
    def ref_pos():
        return np.arange(270).reshape(3, 30, 3)

    @staticmethod
    @pytest.fixture
    def ref_box():
        return np.arange(18).reshape(3, 6)

    def test_single_box(self, ref_pos):
        box = np.array([3, 4, 5, 90, 90, 90])

        mr = MemoryReader(ref_pos, dimensions=box)

        for ts in mr:
            assert_equal(ts.dimensions, box)

    def test_varying_box(self, ref_pos, ref_box):
        mr = MemoryReader(ref_pos, dimensions=ref_box)

        for i, ts in enumerate(mr):
            assert_equal(ts.dimensions, ref_box[i])

    def test_wrong_length(self, ref_pos):
        bad_box = np.arange(12).reshape(2, 6)

        with pytest.raises(ValueError):
            mr = MemoryReader(ref_pos, dimensions=bad_box)

    def test_wrong_shape(self, ref_pos):
        bad_box = np.arange(15).reshape(3, 5)

        with pytest.raises(ValueError):
            mr = MemoryReader(ref_pos, dimensions=bad_box)


class TestMemoryReaderModifications(object):
    # check that modifying MR things behaves as expected
    # in general, modifying the Timestep should be *permanent*
    # this is unlike other Readers!
    n_atoms = 10
    n_frames = 4

    @pytest.fixture()
    def mr_reader(self):
        pos = np.arange(self.n_frames * self.n_atoms * 3).reshape(
            self.n_frames, self.n_atoms, 3
        )
        vel = (
            np.arange(self.n_frames * self.n_atoms * 3).reshape(
                self.n_frames, self.n_atoms, 3
            )
            + 200
        )
        frc = (
            np.arange(self.n_frames * self.n_atoms * 3).reshape(
                self.n_frames, self.n_atoms, 3
            )
            + 400
        )
        box = np.arange(self.n_frames * 6).reshape(self.n_frames, 6) + 600

        return MemoryReader(pos, velocities=vel, forces=frc, dimensions=box)

    @pytest.fixture()
    def mr_universe(self, mr_reader):
        u = mda.Universe.empty(self.n_atoms)
        u.trajectory = mr_reader

        return u

    @pytest.mark.parametrize(
        "attr", ["positions", "velocities", "forces", "dimensions"]
    )
    def test_copying(self, mr_reader, attr):
        mr2 = mr_reader.copy()
        # update the attribute
        ts = mr2.ts
        setattr(ts, attr, 7)
        # check the change worked
        assert_almost_equal(getattr(ts, attr), 7)
        assert ts.positions.shape == (self.n_atoms, 3)
        assert ts.velocities.shape == (self.n_atoms, 3)
        assert ts.forces.shape == (self.n_atoms, 3)
        assert ts.dimensions.shape == (6,)
        # move the Reader around, forcing updates of ts
        ts = mr2[2]
        ts = mr2[0]
        # check our old change is still there
        assert_almost_equal(getattr(ts, attr), 7)

    @pytest.mark.parametrize(
        "attr", ["positions", "velocities", "forces", "dimensions"]
    )
    def test_attr_set(self, mr_universe, attr):
        # same as above, but via a Universe/AtomGroup
        u = mr_universe
        ts = u.trajectory[0]

        setattr(ts, attr, 7)

        assert_almost_equal(getattr(ts, attr), 7)

        ts = u.trajectory[2]
        ts = u.trajectory[0]

        assert_almost_equal(getattr(ts, attr), 7)
        assert u.atoms.positions.shape == (self.n_atoms, 3)
        assert u.atoms.velocities.shape == (self.n_atoms, 3)
        assert u.atoms.forces.shape == (self.n_atoms, 3)
        assert u.atoms.dimensions.shape == (6,)

    @pytest.mark.parametrize("attr", ["velocities", "forces", "dimensions"])
    def test_non_numpy_arr(self, attr):
        with pytest.raises(TypeError):
            mr = MemoryReader(np.zeros((10, 30, 3)), **{attr: "not an array"})
